# -*- coding: utf-8 -*-

"""
This module contains functions to manipulate terminal output.
"""

import typing as T
import blessed
from datetime import datetime, timezone

terminal = blessed.Terminal()
"""
The terminal object is a singleton instance of :class:`blessed.Terminal`.
We only create this object once and use it in the 
:func:`~aws_resource_search.ui.main.run_ui`.
"""

term = terminal


def format_shortcut(key: str) -> str:
    """
    Format a keyboard shortcut key. In this project, the color is magenta.
    Example:

        :magenta:`Enter` to open url
    """
    return f"{term.magenta}{key}{term.normal}"


def highlight_text(text: str) -> str:
    """
    Highlight a text with terminal color. In this project, the color is cyan.
    Example:

        this is a very :cyan:`Important message`!
    """
    return f"{term.cyan}{text}{term.normal}"


def format_resource_type(resource_type: str) -> str:
    """
    Format a resource type text with terminal color. In this project,
    the color is blue. Example:

        :blue:`sfn-statemachine`: name = CognitoUserManagement
    """
    return f"{term.green}{resource_type}{term.normal}"


def format_key(key: str) -> str:
    """
    Format a key of key-value-pair text with terminal color. In this project,
    the color is cyan. Example:

        tag :cyan:`environment` = :yellow:`production`
    """
    return highlight_text(key)


def format_value(value: str) -> str:
    """
    Format a value of key-value-pair text with terminal color. In this project,
    the color is yellow. Example:

        tag :cyan:`environment` = :yellow:`production`
    """
    if isinstance(value, datetime):
        value = str(value.astimezone(tz=timezone.utc).replace(microsecond=0))[:19]
    return f"{term.yellow}{value}{term.normal}"


def format_key_value(key: str, value: T.Any) -> str:
    """
    Format a key-value pair text with terminal color. In this project, key is
    in cyan and value is in yellow. Example:

        tag :cyan:`environment` = :yellow:`production`
    """
    if isinstance(value, datetime):
        value = str(value.astimezone(tz=timezone.utc).replace(microsecond=0))[:19]
    return f"{format_key(key)} = {format_value(value)}"


def remove_text_format(text: str) -> str:
    """
    Remove the terminal format from the given text.
    """
    formats = [
        term.cyan,
        term.yellow,
        term.magenta,
        term.blue,
        term.green,
        term.normal,
    ]
    for c in formats:
        text = text.replace(c, "")
    return text


class ShortcutEnum:
    """
    Formatted keyboard shortcuts::

        print("Tap {ShortcutEnum.ENTER} to open url")

    For example, the above code will print:

        Tap :magenta:`Enter` to open url
    """

    TAB = format_shortcut("Tab")
    ENTER = format_shortcut("Enter")
    CTRL_A = format_shortcut("Ctrl A")
    CTRL_W = format_shortcut("Ctrl W")
    CTRL_U = format_shortcut("Ctrl U")
    CTRL_P = format_shortcut("Ctrl P")
    F1 = format_shortcut("F1")


SUBTITLE = (
    f"🌐 {ShortcutEnum.ENTER} to open url, "
    f"📋 {ShortcutEnum.CTRL_A} to copy arn, "
    f"🔗 {ShortcutEnum.CTRL_U} to copy url, "
    f"👀 {ShortcutEnum.CTRL_P} to view details."
)
"""
The subtitle in the zelfred UI.

The default subtitle is the help text to show the user how to interact with the UI.

Example:

🌐 :magenta:`Enter` to open url, 📋 :magenta:`Ctrl A` to copy arn, 🔗 :magenta:`Ctrl U` to copy url, 👀 :magenta:`Ctrl P` to view details."
"""

SHORT_SUBTITLE = (
    f"🌐 {ShortcutEnum.ENTER}, "
    f"📋 {ShortcutEnum.CTRL_A}, "
    f"🔗 {ShortcutEnum.CTRL_U}, "
    f"👀 {ShortcutEnum.CTRL_P}."
)
"""
A shorter version of subtitle.

Example:

🌐 :magenta:`Enter`, 📋 :magenta:`Ctrl A`, 🔗 :magenta:`Ctrl U`, 👀 :magenta:`Ctrl P`."
"""



# -*- coding: utf-8 -*-

"""
This module defines the search patterns for those resource types that requires
special handling.
"""

import typing as T

from .compat import TypedDict, cached_property
from .searcher_enum import SearcherEnum

if T.TYPE_CHECKING:
    from .ars_def import ARS


class T_SEARCH_PATTERN(TypedDict):
    partitioner_resource_type: str
    get_boto_kwargs: T.Callable


K_PARTITIONER_RESOURCE_TYPE = "partitioner_resource_type"
K_GET_BOTO_KWARGS = "get_boto_kwargs"


class ArsSearchPatternsMixin:
    """
    todo: docstring
    """
    def get_search_patterns(self: "ARS"):
        """
        This variable defines those resource types that requires a parent resource name
        for the boto3 API call. For example:

        - in order to search glue table, you need to specify glue database
        - in order to search glue job run, you need to specify glue job
        """
        return {
            SearcherEnum.ecr_repository_image.value: {
                K_PARTITIONER_RESOURCE_TYPE: SearcherEnum.ecr_repository.value,
                K_GET_BOTO_KWARGS: lambda partitioner_query: {
                    "repositoryName": partitioner_query
                },
            },
            SearcherEnum.ecs_task_run.value: {
                K_PARTITIONER_RESOURCE_TYPE: SearcherEnum.ecs_cluster.value,
                K_GET_BOTO_KWARGS: lambda partitioner_query: {
                    "cluster": partitioner_query
                },
            },
            SearcherEnum.glue_database_table.value: {
                K_PARTITIONER_RESOURCE_TYPE: SearcherEnum.glue_database.value,
                K_GET_BOTO_KWARGS: lambda partitioner_query: {
                    "DatabaseName": partitioner_query
                },
            },
            SearcherEnum.glue_job_run.value: {
                K_PARTITIONER_RESOURCE_TYPE: SearcherEnum.glue_job.value,
                K_GET_BOTO_KWARGS: lambda partitioner_query: {
                    "JobName": partitioner_query
                },
            },
            SearcherEnum.sfn_state_machine_execution: {
                K_PARTITIONER_RESOURCE_TYPE: SearcherEnum.sfn_state_machine.value,
                K_GET_BOTO_KWARGS: lambda partitioner_query: {
                    "stateMachineArn": self.aws_console.step_function.get_state_machine_arn(
                        partitioner_query
                    )
                },
            },
            SearcherEnum.codebuild_job_run.value: {
                K_PARTITIONER_RESOURCE_TYPE: SearcherEnum.codebuild_project.value,
                K_GET_BOTO_KWARGS: lambda partitioner_query: {
                    "projectName": partitioner_query
                },
            },
            SearcherEnum.lambda_function_alias.value: {
                K_PARTITIONER_RESOURCE_TYPE: SearcherEnum.lambda_function.value,
                K_GET_BOTO_KWARGS: lambda partitioner_query: {
                    "FunctionName": partitioner_query
                },
            },
        }

    @cached_property
    def search_patterns(self: "ARS"):
        return self.get_search_patterns()

    def _clear_search_patterns_cache(self: "ARS"):
        """
        Clear the :meth:`ArsSearchPatternsMixin.search_patterns` cache.
        """
        del self.search_patterns

    def has_partitioner(
        self: "ARS",
        resource_type: str,
    ) -> bool:
        """
        Check if a resource type need a partitioner resource.
        """
        return resource_type in self.search_patterns

    def get_partitioner_resource_type(
        self: "ARS",
        resource_type: str,
    ) -> str:
        """
        Get the partitioner "resource type" of a resource type.
        """
        return self.search_patterns[resource_type][K_PARTITIONER_RESOURCE_TYPE]

    def get_partitioner_boto_kwargs(
        self: "ARS",
        resource_type: str,
        partitioner_query: str,
    ) -> dict:
        """
        Get the boto3 kwargs for the partitioner resource.
        """
        return self.search_patterns[resource_type][K_GET_BOTO_KWARGS](partitioner_query)



# -*- coding: utf-8 -*-

"""
See :class:`SearcherMetadata`.
"""

import dataclasses

from .base_model import BaseModel


@dataclasses.dataclass
class SearcherMetadata(BaseModel):
    """
    This class is a data container for the metadata of a
    :class:`~aws_resource_search.base_searcher.BaseSearcher``.
    It describes the searcher's ID, description, and the path to the
    corresponding searcher module. It doesn't contain any data about
    the searcher's logic.

    This class is the data model for the ``aws_resource_search/searchers.json``

    This class is used heavily in the code generation process.

    .. seealso::

        :ref:`what-is-searcher`

    Example:

    .. code-block:: python

        SearcherMetadata(
            id="s3-bucket",
            desc="A bucket is a container for objects.",
            klass="S3BucketSearcher",
            module="s3",
            ngram="simple storage service",
            var="s3_bucket_searcher"
        )

    :param id: the maintainer defined unique identifier of the resource type,
        using hyphen-case.
    :param desc: the maintainer defined description for the specific resource type.
    :param ngram: ngram search string for the resource type.
    :param module: the path to the searcher module, for example,
        :mod:`/path/to/aws_resource_search/res/awslambda.py (aws_resource_search.res.awslambda) <aws_resource_search.res.awslambda>`
    :param klass: the class name in the searcher module, for example ``LambdaFunction``
    :param var: the variable name of the searcher class instance of the
        searcher module, for example ``lambda_function_searcher``.
    """

    id: str = dataclasses.field()
    desc: str = dataclasses.field()
    ngram: str = dataclasses.field()
    module: str = dataclasses.field(default=None)
    klass: str = dataclasses.field(default=None)
    var: str = dataclasses.field(default=None)

    @property
    def id_snake(self) -> str:
        """
        Example::

            "s3_bucket"
        """
        return self.id.replace("-", "_")

    @property
    def id_slug(self) -> str:
        """
        Example::

            "s3-bucket"
        """
        return self.id.replace("_", "-")



# -*- coding: utf-8 -*-

"""
See :func:`show_aws_info_handler`.
"""

import typing as T

from .. import res_lib as rl
from ..terminal import ShortcutEnum

if T.TYPE_CHECKING:  # pragma: no cover
    from ..ui_def import UI


def show_aws_info_handler(
    ui: "UI",
    line_input: str,
    skip_ui: bool = False,
) -> T.List[rl.ShowAwsInfoItem]:  # pragma: no cover
    """
    Show current AWS account information of the current boto session.

    :param line_input: the original query input by user before the ``!?``.
    """
    ui.render.prompt = f"(Query)"
    items = [
        rl.ShowAwsInfoItem(
            title="📝 See AWS Account Info of current boto session below",
            subtitle=(
                f"Hit {ShortcutEnum.TAB} or {ShortcutEnum.ENTER} to go back, "
                f"Hit {ShortcutEnum.CTRL_W} to pick another AWS profile"
            ),
            uid=f"info-0",
            autocomplete=line_input,
        )
    ]
    bsm = ui.ars.bsm
    for key in [
        "aws_account_alias",
        "aws_account_id",
        "aws_region",
    ]:
        try:
            value = getattr(bsm, key)
        except Exception as e:
            value = f"Unable to get, error: {e}"
        item = rl.ShowAwsInfoItem.from_key_value(
            key=key,
            value=value,
            autocomplete=line_input,
        )
        items.append(item)
    return items



# -*- coding: utf-8 -*-

"""
See :class:`AwsResourceTypeItem`.
"""

import typing as T
import dataclasses

from ..terminal import ShortcutEnum, format_resource_type
from ..compat import TypedDict

from .base_item import BaseArsItem

if T.TYPE_CHECKING:  # pragma: no cover
    from ..documents.resource_type_document import ResourceTypeDocument
    from ..ars_def import ARS
    from ..ui_def import UI


class T_AWS_RESOURCE_TYPE_ITEM_VARIABLES(TypedDict):
    """
    Type hint for the "variables" field in :class:`AwsResourceTypeItem`.
    """

    doc: "ResourceTypeDocument"
    resource_type: str
    console_url: T.Optional[str]


@dataclasses.dataclass
class AwsResourceTypeItem(BaseArsItem):
    """
    Represent an item in the resource type search result.
    """

    variables: T_AWS_RESOURCE_TYPE_ITEM_VARIABLES = dataclasses.field(
        default_factory=dict
    )

    @classmethod
    def from_document(
        cls,
        doc: "ResourceTypeDocument",
        ars: "ARS",
    ):
        """
        For example: if the doc is::

            {"id": "s3-bucket", "name": "s3-bucket"}

        Then the item will be::

            {
                "title": "s3-bucket",
                "subtitle": "hit 'Tab' and enter your query to search.",
                "uid": "s3-bucket",
                "arg": "s3-bucket",
                "autocomplete": "s3-bucket: ",
            }
        """
        resource_type = doc.name
        desc = doc.desc
        searcher = ars.searcher_finder.import_searcher(resource_type)
        try:
            console_url = searcher.doc_class.get_list_resources_console_url(
                console=ars.aws_console
            )
            subtitle = (
                f"hit {ShortcutEnum.TAB} to search, "
                f"hit {ShortcutEnum.ENTER} to list resources in AWS console."
            )
        except NotImplementedError:
            console_url = None
            subtitle = f"hit {ShortcutEnum.TAB} to search"
        return cls(
            title=f"{format_resource_type(resource_type)}: {desc}",
            subtitle=subtitle,
            uid=doc.id,
            arg=resource_type,
            autocomplete=resource_type + ": ",
            variables={
                "doc": doc,
                "resource_type": resource_type,
                "console_url": console_url,
            },
        )

    @classmethod
    def from_many_document(
        cls,
        docs: T.Iterable["ResourceTypeDocument"],
        ars: "ARS",
    ):
        return [cls.from_document(doc, ars) for doc in docs]

    def enter_handler(self, ui: "UI"):  # pragma: no cover
        """
        Behavior:

        For resource type search, when user hit "Enter", it opens the
        AWS console to list resources of this type.
        """
        if self.variables.get("console_url"):
            self.open_url_or_print(ui, self.variables["console_url"])



# -*- coding: utf-8 -*-

"""
Define the config class data model.
"""

import typing as T
import json
import dataclasses
from pathlib import Path

from ..vendor.json_utils import strip_comments
from ..vendor.better_dataclasses import DataClass
from ..vendor.hierarchy_config import SHARED, apply_shared_value

from ..paths import path_config_json
from ..searcher_enum import SearcherEnum


@dataclasses.dataclass
class Resource(DataClass):
    cache_expire: int = dataclasses.field()


@dataclasses.dataclass
class Config(DataClass):
    res: T.Dict[str, Resource] = Resource.map_of_nested_field(default_factory=dict)

    @classmethod
    def load(cls, path: Path = path_config_json) -> "Config":
        if path.exists():
            data = json.loads(strip_comments(path.read_text()))
            apply_shared_value(data)
            return cls.from_dict(data)
        else:
            default_cache_expire = 24 * 60 * 60

            default_data = {
                SHARED: {
                    "res.*.cache_expire": default_cache_expire,
                },
                "res": {
                    res_type.value: {
                        "cache_expire": default_cache_expire,
                    }
                    for res_type in SearcherEnum
                },
            }
            res = default_data["res"]
            res[SearcherEnum.codebuild_job_run.value]["cache_expire"] = 5 * 60
            res[SearcherEnum.ecs_task_run.value]["cache_expire"] = 5 * 60
            res[SearcherEnum.glue_job_run.value]["cache_expire"] = 5 * 60
            res[SearcherEnum.sfn_state_machine_execution.value]["cache_expire"] = 5 * 60
            path_config_json.write_text(json.dumps(default_data, indent=4))
            return cls.load(path=path)

    def get_cache_expire(self, res_type: str) -> int:
        return self.res[res_type].cache_expire



# -*- coding: utf-8 -*-

"""
See :class:`BaseArsDocument`.
"""

import typing as T
import dataclasses

from ..base_model import BaseModel


@dataclasses.dataclass
class BaseArsDocument(BaseModel):
    """
    """


T_ARS_DOCUMENT = T.TypeVar("T_ARS_DOCUMENT", bound=BaseArsDocument)



# -*- coding: utf-8 -*-

import typing as T
import dataclasses
from datetime import datetime

import sayt.api as sayt
import aws_arns.api as arns
import aws_console_url.api as acu

from .. import res_lib as rl

if T.TYPE_CHECKING:
    from ..ars_def import ARS


class GlueMixin:
    @property
    def description(self: rl.ResourceDocument) -> str:
        return rl.get_description(self.raw_data, "Description")


@dataclasses.dataclass
class GlueDatabase(rl.ResourceDocument, GlueMixin):
    # fmt: off
    database_arn: str = dataclasses.field(metadata={"field": sayt.StoredField(name="database_arn")})
    # fmt: on

    @property
    def catalog_id(self) -> str:
        return self.raw_data["CatalogId"]

    @property
    def database(self) -> str:
        return self.raw_data["Name"]

    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        return cls(
            raw_data=resource,
            id=resource["Name"],
            name=resource["Name"],
            database_arn="arn:aws:glue:{aws_region}:{aws_account_id}:database/{database}".format(
                aws_region=bsm.aws_region,
                aws_account_id=bsm.aws_account_id,
                database=resource["Name"],
            ),
        )

    @property
    def title(self) -> str:
        return rl.format_key_value("name", self.database)

    @property
    def subtitle(self) -> str:
        return "{}, {}".format(
            self.description,
            self.short_subtitle,
        )

    @property
    def autocomplete(self) -> str:
        return self.name

    @property
    def arn(self) -> str:
        return self.database_arn

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.glue.get_database(database_or_arn=self.arn)

    @classmethod
    def get_list_resources_console_url(cls, console: acu.AWSConsole) -> str:
        return console.glue.databases


class GlueDatabaseSearcher(rl.BaseSearcher[GlueDatabase]):
    pass


glue_database_searcher = GlueDatabaseSearcher(
    # list resources
    service="glue",
    method="get_databases",
    is_paginator=True,
    default_boto_kwargs={"PaginationConfig": {"MaxItems": 9999, "PageSize": 1000}},
    result_path=rl.ResultPath("DatabaseList"),
    # extract document
    doc_class=GlueDatabase,
    # search
    resource_type=rl.SearcherEnum.glue_database.value,
    fields=GlueDatabase.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.glue_database.value),
    more_cache_key=None,
)


@dataclasses.dataclass
class GlueTable(rl.ResourceDocument, GlueMixin):
    # fmt: off
    table_arn: str = dataclasses.field(metadata={"field": sayt.StoredField(name="table_arn")})
    # fmt: on

    @property
    def catalog_id(self) -> str:
        return self.raw_data["CatalogId"]

    @property
    def database(self) -> str:
        return self.raw_data["DatabaseName"]

    @property
    def table(self) -> str:
        return self.raw_data["Name"]

    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        return cls(
            raw_data=resource,
            id="{}.{}".format(resource["DatabaseName"], resource["Name"]),
            name="{}.{}".format(resource["DatabaseName"], resource["Name"]),
            table_arn="arn:aws:glue:{aws_region}:{aws_account_id}:table/{database}/{table}".format(
                aws_region=bsm.aws_region,
                aws_account_id=bsm.aws_account_id,
                database=resource["DatabaseName"],
                table=resource["Name"],
            ),
        )

    @property
    def title(self) -> str:
        return rl.format_key_value("fullname", f"{self.database}.{self.table}")

    @property
    def subtitle(self) -> str:
        return "{}, {}".format(
            self.description,
            self.short_subtitle,
        )

    @property
    def autocomplete(self) -> str:
        return f"{self.database}@{self.table}"

    @property
    def arn(self) -> str:
        return self.table_arn

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.glue.get_table(table_or_arn=self.arn)

    @classmethod
    def get_list_resources_console_url(cls, console: acu.AWSConsole) -> str:
        return console.glue.tables


class GlueTableSearcher(rl.BaseSearcher[GlueTable]):
    pass


glue_table_searcher = GlueTableSearcher(
    # list resources
    service="glue",
    method="get_tables",
    is_paginator=True,
    default_boto_kwargs={"PaginationConfig": {"MaxItems": 9999, "PageSize": 1000}},
    result_path=rl.ResultPath("TableList"),
    # extract document
    doc_class=GlueTable,
    # search
    resource_type=rl.SearcherEnum.glue_database_table.value,
    fields=GlueTable.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.glue_database_table.value),
    more_cache_key=lambda boto_kwargs: [boto_kwargs["DatabaseName"]],
)


@dataclasses.dataclass
class GlueJob(rl.ResourceDocument, GlueMixin):
    # fmt: off
    job_arn: str = dataclasses.field(metadata={"field": sayt.StoredField(name="job_arn")})
    # fmt: on

    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        return cls(
            raw_data=resource,
            id=resource["Name"],
            name=resource["Name"],
            job_arn="arn:aws:glue:{aws_region}:{aws_account_id}:job/{job}".format(
                aws_region=bsm.aws_region,
                aws_account_id=bsm.aws_account_id,
                job=resource["Name"],
            ),
        )

    @property
    def title(self) -> str:
        return rl.format_key_value("name", self.name)

    @property
    def subtitle(self) -> str:
        return "{}, {}".format(
            self.description,
            self.short_subtitle,
        )

    @property
    def autocomplete(self) -> str:
        return self.name

    @property
    def arn(self) -> str:
        return self.job_arn

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.glue.get_job(name_or_arn=self.arn)

    @classmethod
    def get_list_resources_console_url(cls, console: acu.AWSConsole) -> str:
        return console.glue.jobs

    # fmt: off
    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        from_detail = rl.DetailItem.from_detail
        url = self.get_console_url(console=ars.aws_console)
        detail_items = rl.DetailItem.get_initial_detail_items(doc=self, ars=ars)

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.glue_client.get_job(JobName=self.name)
            job_dct = res["Job"]
            description = job_dct.get("Description", "NA")
            role_arn = job_dct["Role"]
            glue_version = job_dct.get("GlueVersion", "NA")
            worker_type = job_dct.get("WorkerType", "NA")
            number_of_workers = job_dct.get("NumberOfWorkers", "NA")
            max_concurrent_runs = job_dct.get("ExecutionProperty", {}).get("MaxConcurrentRuns", "unknown")
            max_retries = job_dct.get("MaxRetries", "NA")
            execution_class = job_dct.get("ExecutionClass", "NA")
            script_location = job_dct.get("Command", {}).get("ScriptLocation", "NA")

            detail_items.extend([
                from_detail("description", description, url=url),
                from_detail("role_arn", role_arn, url=ars.aws_console.iam.get_role(role_arn)),
                from_detail("glue_version", glue_version, url=url),
                from_detail("worker_type", worker_type, url=url),
                from_detail("number_of_workers", number_of_workers, url=url),
                from_detail("max_concurrent_runs", max_concurrent_runs, url=url),
                from_detail("max_retries", max_retries, url=url),
                from_detail("execution_class", execution_class, url=url),
                from_detail("script_location", script_location, url=ars.aws_console.s3.get_console_url(uri_liked=script_location)),
            ])

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.glue_client.get_tags(ResourceArn=self.arn)
            tags = rl.extract_tags(res)
            detail_items.extend(rl.DetailItem.from_tags(tags, url))

        return detail_items
    # fmt: on


class GlueJobSearcher(rl.BaseSearcher[GlueJob]):
    pass


glue_job_searcher = GlueJobSearcher(
    # list resources
    service="glue",
    method="get_jobs",
    is_paginator=True,
    default_boto_kwargs={"PaginationConfig": {"MaxItems": 9999, "PageSize": 1000}},
    result_path=rl.ResultPath("Jobs"),
    # extract document
    doc_class=GlueJob,
    # search
    resource_type=rl.SearcherEnum.glue_job.value,
    fields=GlueJob.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.glue_job.value),
    more_cache_key=None,
)


glue_job_run_state_icon_mapper = {
    "STARTING": "🟡",
    "RUNNING": "🔵",
    "STOPPING": "🟡",
    "STOPPED": "⚫",
    "SUCCEEDED": "🟢",
    "FAILED": "🔴",
    "TIMEOUT": "🔴",
    "ERROR": "🔴",
    "WAITING'": "🟡",
}


@dataclasses.dataclass
class GlueJobRun(rl.ResourceDocument):
    # fmt: off
    started_on: datetime = dataclasses.field(metadata={"field": sayt.DatetimeField(name="started_on", sortable=True, ascending=False, stored=True)})
    # fmt: on

    @property
    def job_name(self) -> str:
        return self.raw_data["JobName"]

    @property
    def state(self) -> str:
        return self.raw_data["JobRunState"]

    @property
    def completed_on(self) -> datetime:
        return self.raw_data.get("CompletedOn")

    @property
    def execution_time(self) -> int:
        return self.raw_data.get("ExecutionTime")

    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        return cls(
            raw_data=resource,
            started_on=resource.get("StartedOn"),
            id=resource["Id"],
            name=resource["Id"],
        )

    @property
    def title(self) -> str:
        return rl.format_key_value("job_run_id", self.id)

    @property
    def subtitle(self) -> str:
        state_icon = glue_job_run_state_icon_mapper[self.state]
        return "{} | {} | {} | {}, {}".format(
            f"{state_icon} {self.state}",
            rl.format_key_value("start", str(self.started_on)[:19]),
            rl.format_key_value("end", str(self.completed_on)[:19]),
            rl.format_key_value("elapsed", f"{self.execution_time} secs"),
            self.short_subtitle,
        )

    @property
    def autocomplete(self) -> str:
        return f"{self.job_name}@{self.id}"

    @property
    def arn(self) -> str:
        raise NotImplementedError("Glue Job Run does not have an ARN")

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.glue.get_glue_job_run(
            job_name_or_arn=self.job_name, job_run_id=self.id
        )

    # fmt: off
    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        from_detail = rl.DetailItem.from_detail
        url = self.get_console_url(console=ars.aws_console)
        detail_items = [
            from_detail("job_run_id", self.id, url=url),
        ]

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.glue_client.get_job_run(
                JobName=self.job_name,
                RunId=self.id,
            )
            dct = res["JobRun"]

            error_message = dct.get("ErrorMessage", "NA")
            log_group_name = dct.get("LogGroupName", "NA")

            detail_items.extend([
                from_detail("error_message", error_message, url=url),
                from_detail("output_logs", log_group_name, url=ars.aws_console.cloudwatch.get_log_stream(stream_name_or_arn=self.id, group_name=f"{log_group_name}/output")),
                from_detail("error_logs", log_group_name, url=ars.aws_console.cloudwatch.get_log_stream(stream_name_or_arn=self.id, group_name=f"{log_group_name}/error")),
            ])

            args = dct.get("Arguments", {})
            detail_items.extend(
                [
                    rl.DetailItem.new(
                        title=f"📝 arg: {rl.format_key_value(k, v)}",
                        subtitle=f"🌐 {rl.ShortcutEnum.ENTER} to open url, 📋 {rl.ShortcutEnum.CTRL_A} to copy arg value.",
                        uid=f"arg {k}",
                        copy=v,
                        url=url,
                    )
                    for k, v in args.items()
                ]
            )
            if len(args) == 0:
                detail_items.append(
                    rl.DetailItem.new(
                        title=f"📝 arg: 🔴 No arg found",
                        subtitle=f"no arg found",
                        uid=f"no arg found",
                        url=url,
                    )
                )

        return detail_items
    # fmt: on


class GlueJobRunSearcher(rl.BaseSearcher[GlueJobRun]):
    pass


glue_job_run_searcher = GlueJobRunSearcher(
    # list resources
    service="glue",
    method="get_job_runs",
    is_paginator=True,
    default_boto_kwargs={"PaginationConfig": {"MaxItems": 9999, "PageSize": 200}},
    result_path=rl.ResultPath("JobRuns"),
    # extract document
    doc_class=GlueJobRun,
    # search
    resource_type=rl.SearcherEnum.glue_job_run.value,
    fields=GlueJobRun.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.glue_job_run.value),
    more_cache_key=lambda boto_kwargs: [boto_kwargs["JobName"]],
)

glue_crawler_state_icon_mapper = {
    "READY": "🔵",
    "RUNNING": "🟢",
    "STOPPING": "🔴",
}


@dataclasses.dataclass
class GlueCrawler(rl.ResourceDocument, GlueMixin):
    # fmt: off
    crawler_arn: str = dataclasses.field(metadata={"field": sayt.StoredField(name="crawler_arn")})
    # fmt: on

    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        return cls(
            raw_data=resource,
            id=resource["Name"],
            name=resource["Name"],
            crawler_arn=arns.res.GlueCrawler.new(
                aws_account_id=bsm.aws_account_id,
                aws_region=bsm.aws_region,
                name=resource["Name"],
            ).to_arn(),
        )

    @property
    def title(self) -> str:
        return rl.format_key_value("name", self.name)

    @property
    def autocomplete(self) -> str:
        return self.name

    @property
    def arn(self) -> str:
        return self.crawler_arn

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.glue.get_crawler(name_or_arn=self.arn)

    @classmethod
    def get_list_resources_console_url(cls, console: acu.AWSConsole) -> str:
        return console.glue.crawlers

    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        from_detail = rl.DetailItem.from_detail
        url = self.get_console_url(console=ars.aws_console)
        detail_items = rl.DetailItem.get_initial_detail_items(doc=self, ars=ars)

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.glue_client.get_crawler(Name=self.name)
            dct = res["Crawler"]

            description = dct.get("Description", "NA")
            state = dct.get("State", "NA")
            role_arn = dct["Role"]
            if not role_arn.startswith("arn:"):
                role_arn = arns.res.IamRole.new(
                    aws_account_id=ars.bsm.aws_account_id,
                    name=role_arn,
                ).to_arn()

            state_icon = glue_crawler_state_icon_mapper[state]
            # fmt: off
            detail_items.extend([
                from_detail("description", description, url=url),
                from_detail("state", state, f"{state_icon} {state}", url=url),
                from_detail("role_arn", role_arn, url=ars.aws_console.iam.get_role(role_arn)),
            ])
            # fmt: on

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.glue_client.get_tags(ResourceArn=self.arn)
            tags = rl.extract_tags(res)
            detail_items.extend(rl.DetailItem.from_tags(tags, url))

        return detail_items


class GlueCrawlerSearcher(rl.BaseSearcher[GlueCrawler]):
    pass


glue_crawler_searcher = GlueCrawlerSearcher(
    # list resources
    service="glue",
    method="get_crawlers",
    is_paginator=True,
    default_boto_kwargs={"PaginationConfig": {"MaxItems": 9999, "PageSize": 1000}},
    result_path=rl.ResultPath("Crawlers"),
    # extract document
    doc_class=GlueCrawler,
    # search
    resource_type=rl.SearcherEnum.glue_crawler.value,
    fields=GlueCrawler.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.glue_crawler.value),
    more_cache_key=None,
)



# -*- coding: utf-8 -*-

import typing as T
import dataclasses
from datetime import datetime

import aws_console_url.api as acu
import sayt.api as sayt

from .. import res_lib as rl

if T.TYPE_CHECKING:
    from ..ars_def import ARS


sfn_statemachine_status_icon_mapper = {
    "ACTIVE": "🟢",
    "DELETING": "🔴",
}

sfn_statemachine_type_icon_mapper = {
    "STANDARD": "⌛",
    "EXPRESS": "🚀",
}


@dataclasses.dataclass
class SfnStateMachine(rl.ResourceDocument):
    # fmt: off
    type: str = dataclasses.field(metadata={"field": sayt.NgramWordsField(name="type", minsize=2, maxsize=4, stored=True)})
    # fmt: on

    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        return cls(
            raw_data=resource,
            id=resource["name"],
            name=resource["name"],
            type=resource.get("type", "STANDARD"),
        )

    @property
    def state_machine_name(self) -> str:
        return self.name

    @property
    def title(self) -> str:
        return rl.format_key_value("name", self.name)

    @property
    def subtitle(self) -> str:
        type_icon = sfn_statemachine_type_icon_mapper[self.type]
        return "{}, {}".format(
            rl.format_key_value("type", f"{type_icon} {self.type}"),
            self.short_subtitle,
        )

    @property
    def autocomplete(self) -> str:
        return self.name

    @property
    def arn(self) -> str:
        return self.raw_data["stateMachineArn"]

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.step_function.get_state_machine_view_tab(name_or_arn=self.arn)

    @classmethod
    def get_list_resources_console_url(cls, console: acu.AWSConsole) -> str:
        return console.step_function.state_machines

    # fmt: off
    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        from_detail = rl.DetailItem.from_detail
        url = self.get_console_url(console=ars.aws_console)
        detail_items = rl.DetailItem.get_initial_detail_items(doc=self, ars=ars, arn_key="statemachine_arn")

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.sfn_client.describe_state_machine(stateMachineArn=self.arn)
            status = res["status"]
            role_arn = res["roleArn"]
            definition = res["definition"]
            type = res.get("type", "STANDARD")
            creation_date = res["creationDate"]

            status_icon = sfn_statemachine_status_icon_mapper[status]
            type_icon = sfn_statemachine_type_icon_mapper[type]
            detail_items.extend([
                from_detail("status", status, value_text=f"{status_icon} {status}", url=url),
                from_detail("🧢 role_arn", role_arn, url=ars.aws_console.iam.get_role(role_arn)),
                from_detail("definition", definition, value_text=self.one_line(definition), url=url),
                from_detail("type", type, value_text=f"{type_icon} {type}", url=url),
                from_detail("creation_date", creation_date, url=url),
            ])

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.sfn_client.list_tags_for_resource(resourceArn=self.arn)
            tags = rl.extract_tags(res)
            detail_items.extend(rl.DetailItem.from_tags(tags, url))

        return detail_items
    # fmt: on


class SfnStateMachineSearcher(rl.BaseSearcher[SfnStateMachine]):
    pass


sfn_state_machine_searcher = SfnStateMachineSearcher(
    # list resources
    service="stepfunctions",
    method="list_state_machines",
    is_paginator=True,
    default_boto_kwargs={"PaginationConfig": {"MaxItems": 9999, "PageSize": 1000}},
    result_path=rl.ResultPath("stateMachines"),
    # extract document
    doc_class=SfnStateMachine,
    # search
    resource_type=rl.SearcherEnum.sfn_state_machine.value,
    fields=SfnStateMachine.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.sfn_state_machine.value),
    more_cache_key=None,
)


sfn_execution_status_icon_mapper = {
    "RUNNING": "🔵",
    "SUCCEEDED": "🟢",
    "FAILED": "🔴",
    "TIMED_OUT": "🔴",
    "ABORTED'": "⚫",
}


@dataclasses.dataclass
class SfnExecution(rl.ResourceDocument):
    # fmt: off
    status: str = dataclasses.field(metadata={"field": sayt.NgramWordsField(name="status", minsize=2, maxsize=4, stored=True)})
    start_at: datetime = dataclasses.field(metadata={"field": sayt.DatetimeField(name="start_at", sortable=True, ascending=False, stored=True)})
    # fmt: on

    @property
    def state_machine_name(self) -> str:
        return self.raw_data["stateMachineArn"].split(":")[-1]

    @property
    def end_at(self) -> datetime:
        return rl.get_datetime(self.raw_data, "stopDate")

    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        return cls(
            raw_data=resource,
            id=resource["executionArn"].split(":")[-1],
            name=resource["executionArn"].split(":")[-1],
            status=resource["status"],
            start_at=rl.get_datetime(resource, "startDate"),
        )

    @property
    def title(self) -> str:
        return rl.format_key_value("execution_name", self.name)

    @property
    def subtitle(self) -> str:
        status_icon = sfn_execution_status_icon_mapper[self.status]
        return "{}, {}, {}, {}".format(
            f"{status_icon} {self.status}",
            rl.format_key_value("start", self.start_at),
            rl.format_key_value("end", self.end_at),
            self.short_subtitle,
        )

    @property
    def autocomplete(self) -> str:
        return f"{self.state_machine_name}@{self.name}"

    @property
    def arn(self) -> str:
        return self.raw_data["executionArn"]

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.step_function.get_state_machine_execution(
            exec_id_or_arn=self.arn
        )

    # fmt: off
    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        from_detail = rl.DetailItem.from_detail
        url = self.get_console_url(console=ars.aws_console)
        detail_items = rl.DetailItem.get_initial_detail_items(doc=self, ars=ars, arn_key="exec_arn")

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.sfn_client.describe_execution(executionArn=self.arn)
            status = res["status"]
            state_machine_arn = res["stateMachineArn"]
            state_machine_version_arn = res.get("stateMachineVersionArn")
            state_machine_alias_arn = res.get("stateMachineAliasArn")
            input = res["input"]
            output = res.get("output")
            error = res.get("error")
            cause = res.get("cause")

            status_icon = sfn_execution_status_icon_mapper[status]
            detail_items.extend([
                from_detail("status", status, value_text=f"{status_icon} {status}", url=url),
                from_detail("state_machine_arn", state_machine_arn, url=ars.aws_console.step_function.get_state_machine_view_tab(state_machine_arn)),
                from_detail("state_machine_version_arn", state_machine_version_arn, url=url) if state_machine_version_arn else None,
                from_detail("state_machine_alias_arn", state_machine_alias_arn, url=url) if state_machine_alias_arn else None,
                from_detail("input", input, value_text=self.one_line(input), url=url),
                from_detail("output", output, value_text=self.one_line(output), url=url),
                from_detail("error", error, value_text=self.one_line(error), url=url),
                from_detail("cause", cause, value_text=self.one_line(cause), url=url),
            ])

        detail_items = [item for item in detail_items if item is not None]
        return detail_items
    # fmt: on


class SfnExecutionSearcher(rl.BaseSearcher[SfnExecution]):
    pass


sfn_execution_searcher = SfnExecutionSearcher(
    # list resources
    service="stepfunctions",
    method="list_executions",
    is_paginator=True,
    default_boto_kwargs={"PaginationConfig": {"MaxItems": 9999, "PageSize": 1000}},
    result_path=rl.ResultPath("executions"),
    # extract document
    doc_class=SfnExecution,
    # search
    resource_type=rl.SearcherEnum.sfn_state_machine_execution.value,
    fields=SfnExecution.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(
        rl.SearcherEnum.sfn_state_machine_execution.value
    ),
    more_cache_key=lambda boto_kwargs: [boto_kwargs["stateMachineArn"]],
)



# -*- coding: utf-8 -*-

import typing as T
import random

from .utils import guid, envs, rand_env

if T.TYPE_CHECKING:
    from .main import FakeAws


class StepFunctionMixin:
    sfn_state_machines: T.List[str] = list()

    @classmethod
    def create_state_machines(cls: T.Type["FakeAws"]):
        for ith in range(1, 1 + 10):
            env = rand_env()
            name = f"{env}-{guid}-{ith}-sfn-state-machine"
            sm_type = random.choice(["STANDARD", "EXPRESS"])
            cls.bsm.sfn_client.create_state_machine(
                name=name,
                definition="{}",
                roleArn=f"arn:aws:iam::{cls.bsm.aws_account_id}:role/AWSStepFunctionRoleDefault",
                type=sm_type,
            )
            cls.sfn_state_machines.append(name)

            if sm_type == "STANDARD":
                sm_arn = f"arn:aws:states:{cls.bsm.aws_region}:{cls.bsm.aws_account_id}:stateMachine:{name}"
                cls.bsm.sfn_client.start_execution(stateMachineArn=sm_arn, input="{}")



# -*- coding: utf-8 -*-

from aws_resource_search.items.exception_item import ExceptionItem


class TestExceptionItem:
    def test_from_error(self):
        try:
            raise ValueError("something wrong")
        except ValueError as e:
            item = ExceptionItem.from_error(error=e, uid="uid")
            assert __file__ in item.variables["traceback"]
            # ------------------------------------------------------------------
            # We only test this action manually on developer's machine, not in CI.
            # ------------------------------------------------------------------
            # item.enter_handler(ui=None)


if __name__ == "__main__":
    from aws_resource_search.tests.helper import run_cov_test

    run_cov_test(
        __file__,
        "aws_resource_search.items.exception_item",
        preview=False,
    )



# -*- coding: utf-8 -*-

import shutil

from ..paths import dir_aws_resource_search, dir_index, dir_cache


def main():
    print(f"clear index in {dir_index}")
    if dir_index.exists():
        shutil.rmtree(dir_index, ignore_errors=True)
    print(f"clear cache in {dir_cache}")
    if dir_cache.exists():
        shutil.rmtree(dir_cache, ignore_errors=True)
    print(f"done, you can verify at file://{dir_aws_resource_search}")



# -*- coding: utf-8 -*-

"""
See :class:`DetailItem`.
"""

import typing as T
import dataclasses
import contextlib

import zelfred.api as zf

try:
    import pyperclip
except ImportError:  # pragma: no cover
    pass

from ..terminal import ShortcutEnum, format_key_value
from ..compat import TypedDict
from .base_item import BaseArsItem, T_ARS_ITEM
from .exception_item import ExceptionItem

if T.TYPE_CHECKING:  # pragma: no cover
    from ..documents.resource_document import ResourceDocument
    from ..ars_def import ARS


class T_DETAIL_ITEM_VARIABLES(TypedDict):
    """
    .. note::

        Copy is more common than url in detail item.
    """
    copy: T.Optional[str]
    url: T.Optional[str]


@dataclasses.dataclass
class DetailItem(BaseArsItem):
    """
    Represent a detail information of an AWS resource in the dropdown menu.

    **Why this class**

    1. We need to write logic to create ``Item`` in the
        ``aws_resource_search.res.${aws_service}.py`` source code.
        This class provides utility methods to make the code clean and descriptive.
    2. An ``Item`` usually comes with user action. We need to store the
        user action argument in ``variables`` and then implement the
        ``xyz_handler`` function to handle the user action. In this project,
        we have a very clear mind of user action patterns. This class provides
        utility methods to enable pre-defined user action smartly based
        on the input, so we no longer need to implement ``variables`` logics
        and ``xyz_handler`` functions in most cases.
    """

    variables: T_DETAIL_ITEM_VARIABLES = dataclasses.field(default_factory=dict)

    @classmethod
    def new(
        cls,
        title: str,
        subtitle: T.Optional[str] = None,
        copy: T.Optional[str] = None,
        url: T.Optional[str] = None,
        uid: T.Optional[str] = None,
        arg: T.Optional[str] = None,
        autocomplete: T.Optional[str] = None,
    ):
        """
        The factory method to create a new ``ArsBaseItem`` instance.

        :param title: first line of the item. It has a checkbox in front of it to
            indicate whether it is selected.
        :param subtitle: second line of the item.
        :param copy: the text to copy to clipboard.
        :param url: the url to open in browser.
        :param uid: item unique id. The UI use this to distinguish different items.
        :param arg: argument that will be passed to the action.
        :param autocomplete: the text that will be filled in the input box when
            user hits ``TAB`` key.
        """
        kwargs = dict(
            title=title,
            arg=arg,
            autocomplete=autocomplete,
            variables={"url": url, "copy": copy},
        )
        if uid:
            kwargs["uid"] = uid

        # fmt: off
        if subtitle is None:
            if url is not None and copy is not None:
                kwargs[
                    "subtitle"] = f"🌐 {ShortcutEnum.ENTER} to open url, 📋 {ShortcutEnum.CTRL_A} to copy value, 🔗 {ShortcutEnum.CTRL_U} to copy url."
            elif url is not None:
                kwargs[
                    "subtitle"] = f"🌐 {ShortcutEnum.ENTER} to open url, 📋 {ShortcutEnum.CTRL_A} or 🔗 {ShortcutEnum.CTRL_U} to copy url."
            elif copy is not None:
                kwargs["subtitle"] = f"📋 {ShortcutEnum.CTRL_A} or 🔗 {ShortcutEnum.CTRL_U} to copy value"
            elif autocomplete:
                kwargs["subtitle"] = f"{ShortcutEnum.TAB} to auto complete."
            else:
                kwargs["subtitle"] = "No subtitle"
        # fmt: on
        else:
            kwargs["subtitle"] = subtitle
        return cls(**kwargs)

    def enter_handler(self, ui: "zf.UI"):  # pragma: no cover
        """
        Behavior:

        - If we have a URL, it will be opened in the browser or print the url
            to the terminal if the terminal device can not do so.
        - If not, the Enter key will function similarly to the Tab key,
            it will autocomplete.
        """
        if self.variables.get("url"):
            self.open_url_or_print(ui, self.variables["url"])
        elif self.autocomplete:
            ui.line_editor.clear_line()
            ui.line_editor.enter_text(self.autocomplete)
        else:
            pass

    def ctrl_a_handler(self, ui: "zf.UI"):  # pragma: no cover
        """
        Behavior:

        - If we have a copy text, it will copy the text to clipboard or print the text
            to the terminal if the terminal device can not do so.
        - If we don't have copy text but have url, then treat url as copy text.
        """
        if self.variables.get("copy"):
            self.copy_or_print(ui, self.variables["copy"])
        elif self.variables.get("url"):
            self.copy_or_print(ui, self.variables["url"])
        else:
            pass

    def ctrl_u_handler(self, ui: "zf.UI"):  # pragma: no cover
        """
        Behavior:

        - If we have an url, it will copy the url to clipboard or print the url
            to the terminal if the terminal device can not do so.
        - If we don't have URL but have copy text, then treat copy text as url.
        """
        if self.variables.get("url"):
            self.copy_or_print(ui, self.variables["url"])
        elif self.variables.get("copy"):
            self.copy_or_print(ui, self.variables["copy"])
        else:
            pass

    @classmethod
    def from_detail(
        cls,
        key: str,
        value: T.Any,
        key_text: T.Optional[str] = None,
        value_text: T.Optional[str] = None,
        url: T.Optional[str] = None,
        copy: T.Optional[str] = None,
        uid: T.Optional[str] = None,
    ):
        """
        A utility method to create :class`DetailItem` from structured detail information.

        This method will be used in
        :class:`aws_resource_search.searcher.base_document.BaseDocument.get_details`
        method.

        A 'detail' is just a simple key value pair. For example, for S3 bucket.
        the 'bucket_name' is the key and the 'your-bucket-name' is the value.

        :param key: key of the detail, it's the original data
        :param value: value of the detail, it's the original data. This is also
            the text that will be copied to clipboard when user tap `Ctrl + A``.
        :param key_text: the ``${key} = ${value}`` text to display in the UI.
            if not provided, use ``key`` as the text.
        :param value_text: the ``${key} = ${value}`` text to display in the UI.
            if not provided, use ``value`` as the text.
        :param url: if specified, user can hit Enter key to open the url in browser,
            and Ctrl + U to copy it.
        :param uid: this is for uid
        """
        if key_text is None:
            key_text = key
        if value_text is None:
            value_text = value
        if copy is None:
            copy = value
        return cls.new(
            title=format_key_value(key_text, value_text),
            uid=uid,
            copy=str(copy),
            url=url,
        )

    @classmethod
    def from_env_vars(
        cls,
        env_vars: T.Dict[str, str],
        url: T.Optional[str] = None,
    ) -> T.List["DetailItem"]:
        """
        A utility method to create many :class`DetailItem` from
        environment variable key value pairs.
        """
        if env_vars:
            return [
                cls.new(
                    title=f"🎯 env var: {format_key_value(k, v)}",
                    uid=f"env_var-{k}",
                    copy=v,
                    url=url,
                )
                for k, v in env_vars.items()
            ]
        else:
            return [
                cls.new(
                    title=f"🏷 env var: 🔴 No environment variable found.",
                    uid=f"no-environment-variable-found",
                    url=url,
                )
            ]

    @classmethod
    def from_tags(
        cls,
        tags: T.Dict[str, str],
        url: T.Optional[str] = None,
    ) -> T.List["DetailItem"]:
        """
        Create MANY :class:`DetailItem` from AWS resource tag key value pairs.
        """
        if tags:
            return [
                cls.new(
                    title=f"🏷 tag: {format_key_value(k, v)}",
                    uid=f"tag {k}",
                    copy=v,
                    url=url,
                )
                for k, v in tags.items()
            ]
        else:
            return [
                cls.new(
                    title=f"🏷 tag: 🔴 No tag found.",
                    uid=f"no-tag-found",
                    url=url,
                )
            ]

    @classmethod
    def get_initial_detail_items(
        cls,
        doc: "ResourceDocument",
        ars: "ARS",
        arn_key: str = "arn",
    ) -> T.List["DetailItem"]:  # pragma: no cover
        """
        Most AWS resource detail should have one ARN item that user can tap
        "Ctrl A" to copy and tap "Enter" to open url. Only a few AWS resource
        doesn't support ARN (for example, glue job run).

        .. note::

            This method is to simplify the authoring of the
            :meth:`aws_resource_search.documents.resource_document.ResourceDocument.get_details` method.

        Usage example:

            >>> class S3BucketDocument(ResourceDocument):
            ...     def get_details(self, ars: ARS):
            ...         detail_items = DetailItem.get_initial_detail_items(self, ars)
            ...         ...
        """
        try:
            return [
                DetailItem.from_detail(
                    key=arn_key,
                    value=doc.arn,
                    url=doc.get_console_url(ars.aws_console),
                ),
            ]
        # the ResourceDocument.arn and ResourceDocument.get_console_url
        # may raise NotImplementedError
        except NotImplementedError:
            return []

    @staticmethod
    @contextlib.contextmanager
    def error_handling(detail_items: T.List["T_ARS_ITEM"]):
        """
        A context manager to add additional detail items to the list.
        It automatically captures exception and creates
        :class:`~aws_resource_search.items.exception_item.ExceptionItem`
        to explain what went wrong.

        Usage example:

            >>> class S3BucketDocument(ResourceDocument):
            ...     def get_details(self, ars: ARS):
            ...         detail_items = DetailItem.get_initial_detail_items(self, ars)
            ...         with DetailItem.error_handling(detail_items):
            ...             res = ars.bsm.s3_client.get_bucket_policy(...)
            ...             detail_items.append(DetailItem.from_detail(...))

        .. note::

            This method is to simplify the authoring of the
            :meth:`aws_resource_search.documents.resource_document.ResourceDocument.get_details` method.
        """
        try:
            yield None
        except Exception as e:
            detail_items.append(ExceptionItem.from_error(error=e))



# -*- coding: utf-8 -*-

"""
See :class:`BaseModel`.
"""

import typing as T
import dataclasses


@dataclasses.dataclass
class BaseModel:
    """
    The base class for all AWS Resource Search data model.

    .. note::

        I intentionally don't use ``better_dataclasses.DataClass`` as the base class
        here. Because the AWS Resource data container is so heavily used in the
        search and performance does really matter. The ``better_dataclasses.DataClass``
        provides additional features like auto-serialization and deserialization
        when using nested dataclass. But it will slow down the performance.
    """

    @classmethod
    def from_dict(cls, data: T.Dict[str, T.Any]):
        """
        Create a new instance from a dict.
        """
        return cls(**data)

    def to_dict(self) -> T.Dict[str, T.Any]:
        """
        Convert the instance to a dict.
        """
        return dataclasses.asdict(self)



# -*- coding: utf-8 -*-

"""
All AWS resource searchers are defined as variables in the
``aws_resource_search.res.${service}.py`` module. We don't want to import them
manually in the code; instead, we need a lazy-loading mechanism to import them
programmatically based on the resource type.
"""

import typing as T
import json
import importlib
import dataclasses

from .paths import path_searchers_json
from .searcher_metadata import SearcherMetadata

if T.TYPE_CHECKING:  # pragma: no cover
    from .base_searcher import T_SEARCHER


@dataclasses.dataclass
class SearcherFinder:
    """
    This is a helper class to access the implemented searcher metadata,
    and also import the per-aws-resource-type searcher object.
    """

    sm_meta_mapper: T.Dict[str, SearcherMetadata] = dataclasses.field(init=False)
    searcher_cache: T.Dict[str, "T_SEARCHER"] = dataclasses.field(default_factory=dict)

    def reload(self):
        """
        Reload the lookup data from ``aws_resource_search/searchers.json``.
        """
        data = json.loads(path_searchers_json.read_text())
        self.sm_meta_mapper = {
            id: SearcherMetadata(
                id=id,
                desc=dct["desc"],
                ngram=dct["ngram"],
                module=dct["module"],
                var=dct["var"],
            )
            for id, dct in data.items()
        }

    def __post_init__(self):
        self.reload()

    def all_resource_types(self) -> T.List[str]:
        """
        Return all resource types.
        """
        return list(self.sm_meta_mapper.keys())

    def is_valid_resource_type(self, resource_type: str) -> bool:
        """
        Check if the resource type is supported.
        """
        return resource_type in self.sm_meta_mapper

    def import_searcher(self, resource_type: str) -> "T_SEARCHER":
        """
        Import the searcher object by resource type, it uses cache to avoid
        loading the same module multiple times.
        """
        if self.is_valid_resource_type(resource_type):
            if resource_type not in self.searcher_cache:
                mod = searcher_finder.sm_meta_mapper[resource_type].module
                var = searcher_finder.sm_meta_mapper[resource_type].var
                module = importlib.import_module(f"aws_resource_search.res.{mod}")
                self.searcher_cache[resource_type] = getattr(module, var)
            return self.searcher_cache[resource_type]
        else:
            raise ValueError(f"Invalid resource type: {resource_type}")


searcher_finder = SearcherFinder()  # singleton object



# -*- coding: utf-8 -*-

import pytest
import moto

from aws_resource_search.downloader import (
    T_RESULT_DATA,
    ResourceIterproxy,
    ResultPath,
    list_resources,
    extract_tags,
)
from aws_resource_search.tests.mock_test import BaseMockTest


class Test(BaseMockTest):
    mock_list = [
        moto.mock_sts,
        moto.mock_s3,
        moto.mock_iam,
    ]

    @classmethod
    def setup_class_post_hook(cls):
        cls.bsm.s3_client.create_bucket(Bucket="bucket1")
        cls.bsm.s3_client.create_bucket(Bucket="bucket2")
        cls.bsm.iam_client.create_group(GroupName="Group1")
        cls.bsm.iam_client.create_group(GroupName="Group2")

    def test(self):
        res = list_resources(
            bsm=self.bsm,
            service="s3",
            method="list_buckets",
            is_paginator=False,
            boto_kwargs=None,
            result_path=ResultPath(path="Buckets"),
        )
        assert [dct["Name"] for dct in res.all()] == ["bucket1", "bucket2"]

        res = list_resources(
            bsm=self.bsm,
            service="iam",
            method="list_groups",
            is_paginator=True,
            boto_kwargs={
                "PaginationConfig": {
                    "MaxItems": 9999,
                    "PageSize": 1000,
                },
            },
            result_path=ResultPath(path="Groups"),
        )
        assert [dct["GroupName"] for dct in res.all()] == ["Group1", "Group2"]


def test_extract_tags():
    tag_data_1 = {"k1": "v1"}
    tag_data_2 = [{"Key": "k1", "Value": "v1"}]
    tag_data_3 = [{"key": "k1", "value": "v1"}]
    fields = [
        "tags",
        "Tags",
        "TagSet",
    ]
    for field in fields:
        for tag_data in [tag_data_1, tag_data_2, tag_data_3]:
            data = {field: tag_data}
            assert extract_tags(data) == {"k1": "v1"}

    with pytest.raises(ValueError):
        extract_tags({"tags": [{"KeyName": 1}]})

    with pytest.raises(TypeError):
        extract_tags({"tags": 1})


if __name__ == "__main__":
    from aws_resource_search.tests.helper import run_cov_test

    run_cov_test(
        __file__,
        "aws_resource_search.downloader",
        preview=False,
    )



# -*- coding: utf-8 -*-

import typing as T
import json
import random

from .utils import guid, envs, rand_env

if T.TYPE_CHECKING:
    from .main import FakeAws


class LambdaMixin:
    lambda_layers: T.List[str] = list()
    lambda_functions: T.List[str] = list()

    @classmethod
    def create_lambda_layers(
        cls: T.Type["FakeAws"],
        s3_bucket: T.Optional[str] = None,
    ):
        if s3_bucket is None:
            s3_bucket = random.choice(cls.s3_buckets)
        cls.bsm.s3_client.put_object(Bucket=s3_bucket, Key="my-layer.zip", Body=b"")
        for ith in range(1, 1 + 10):
            env = rand_env()
            name = f"{env}-{guid}-{ith}-lbd-layer"
            cls.bsm.lambda_client.publish_layer_version(
                LayerName=name,
                Content=dict(
                    S3Bucket=s3_bucket,
                    S3Key="my-key.zip",
                ),
            )
            cls.lambda_layers.append(name)

    @classmethod
    def create_lambda_functions(
        cls: T.Type["FakeAws"],
        s3_bucket: T.Optional[str] = None,
    ):
        cls.bsm.iam_client.create_role(
            RoleName=f"AwsLambdaDefaultRole",
            AssumeRolePolicyDocument=json.dumps(
                {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Effect": "Allow",
                            "Principal": {"Service": "lambda.amazonaws.com"},
                            "Action": "sts:AssumeRole",
                        }
                    ],
                }
            ),
        )

        if s3_bucket is None:
            s3_bucket = random.choice(cls.s3_buckets)

        cls.bsm.s3_client.put_object(Bucket=s3_bucket, Key="my-source.zip", Body=b"")
        for ith in range(1, 1 + 10):
            env = rand_env()
            name = f"{env}-{guid}-{ith}-lbd-func"
            cls.bsm.lambda_client.create_function(
                FunctionName=name,
                Role="arn:aws:iam::123456789012:role/AwsLambdaDefaultRole",
                Handler="lambda_function.lambda_handler",
                MemorySize=256,
                Timeout=3,
                Code=dict(
                    S3Bucket=s3_bucket,
                    S3Key="my-key.zip",
                ),
            )
            cls.lambda_functions.append(name)



# -*- coding: utf-8 -*-

from aws_resource_search.tests.fake_aws.api import FakeAws


class TestARSDef(FakeAws):
    @classmethod
    def setup_class_post_hook(cls):
        cls.setup_ars()

    def _test_all_resource_types(self):
        _ = self.ars.all_resource_types()

    def _test_is_valid_resource_type(self):
        assert self.ars.is_valid_resource_type("s3-bucket") is True
        assert self.ars.is_valid_resource_type("invalid") is False

    def _test_set_profile(self):
        self.ars.set_profile()

    def test(self):
        self._test_all_resource_types()
        self._test_is_valid_resource_type()
        self._test_set_profile()


if __name__ == "__main__":
    from aws_resource_search.tests.helper import run_cov_test

    run_cov_test(__file__, "aws_resource_search.ars_def", preview=False)



# -*- coding: utf-8 -*-

import typing as T
import dataclasses
from datetime import datetime

import sayt.api as sayt
import aws_console_url.api as acu

from .. import res_lib as rl

if T.TYPE_CHECKING:
    from ..ars_def import ARS


cloudformation_stack_status_icon_mapper = {
    "CREATE_IN_PROGRESS": "🟡",
    "CREATE_FAILED": "🔴",
    "CREATE_COMPLETE": "🟢🟢",
    "ROLLBACK_IN_PROGRESS": "🟡",
    "ROLLBACK_FAILED": "🔴",
    "ROLLBACK_COMPLETE": "🟢🟡",
    "DELETE_IN_PROGRESS": "🟡",
    "DELETE_FAILED": "🔴",
    "DELETE_COMPLETE": "🟢🔴",
    "UPDATE_IN_PROGRESS": "🟡",
    "UPDATE_COMPLETE_CLEANUP_IN_PROGRESS": "🟡",
    "UPDATE_COMPLETE": "🟢🟢",
    "UPDATE_FAILED": "🔴",
    "UPDATE_ROLLBACK_IN_PROGRESS": "🟡",
    "UPDATE_ROLLBACK_FAILED": "🔴",
    "UPDATE_ROLLBACK_COMPLETE_CLEANUP_IN_PROGRESS": "🟡",
    "UPDATE_ROLLBACK_COMPLETE": "🟢🟡",
    "REVIEW_IN_PROGRESS": "🟡",
    "IMPORT_IN_PROGRESS": "🟡",
    "IMPORT_COMPLETE": "🟢🟢",
    "IMPORT_ROLLBACK_IN_PROGRESS": "🟡",
    "IMPORT_ROLLBACK_FAILED": "🔴",
    "IMPORT_ROLLBACK_COMPLETE'": "🟢🟡",
}


@dataclasses.dataclass
class CloudFormationStack(rl.ResourceDocument):
    # fmt: off
    status: str = dataclasses.field(metadata={"field": sayt.NgramWordsField(name="status", minsize=2, maxsize=4, stored=True)})
    last_updated_time: datetime = dataclasses.field(metadata={"field": sayt.DatetimeField(name="last_updated_time", sortable=True, ascending=False, stored=True)})
    # fmt: on

    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        return cls(
            raw_data=resource,
            id=resource["StackName"],
            name=resource["StackName"],
            status=resource["StackStatus"],
            last_updated_time=rl.get_datetime(resource, "LastUpdatedTime"),
        )

    @property
    def title(self) -> str:
        return rl.format_key_value("stack_name", self.name)

    @property
    def subtitle(self) -> str:
        status_icon = cloudformation_stack_status_icon_mapper[self.status]
        return "{}, {}, {}".format(
            f"{status_icon} {self.status}",
            rl.format_key_value(
                "update_at", rl.to_simple_dt_fmt(self.last_updated_time)
            ),
            self.short_subtitle,
        )

    @property
    def autocomplete(self) -> str:
        return self.name

    @property
    def arn(self) -> str:
        return self.raw_data["StackId"]

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.cloudformation.get_stack(name_or_arn=self.arn)

    @classmethod
    def get_list_resources_console_url(cls, console: acu.AWSConsole) -> str:
        return console.cloudformation.stacks

    # fmt: off
    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        from_detail = rl.DetailItem.from_detail
        detail_items = [] # don't use self.get_initial_detail_items here, the arn may change

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.cloudformation_client.describe_stacks(StackName=self.name)
            stacks = res.get("Stacks", [])
            if len(stacks) == 0:
                return [
                    rl.DetailItem.new(
                        title="🚨 Stack not found, maybe it's deleted?",
                        subtitle=f"{rl.ShortcutEnum.ENTER} to verify in AWS Console",
                        url=ars.aws_console.cloudformation.filter_stack(name=self.name),
                    )
                ]
            stack = stacks[0]
            arn = stack["StackId"]
            url = ars.aws_console.cloudformation.get_stack(arn)

            status = stack["StackStatus"]
            role_arn = stack.get("RoleARN", "NA")
            status_icon = cloudformation_stack_status_icon_mapper[self.status]
            detail_items = [
                from_detail("arn", arn, url=url),
                from_detail("status", status, value_text=f"{status_icon} {status}", url=url),
                from_detail("role_arn", role_arn, url=ars.aws_console.iam.get_role(role_arn)),
            ]

            outputs: dict = {dct["OutputKey"]: dct for dct in stack.get("Outputs", [])}
            detail_items.extend(
                [
                    rl.DetailItem.new(
                        title="🎯 output: {} (export = {})".format(
                            rl.format_key_value(k, dct["OutputValue"]),
                            dct.get("ExportName", "NA"),
                        ),
                        subtitle=f"🌐 {rl.ShortcutEnum.ENTER} to open url, 📋 {rl.ShortcutEnum.CTRL_A} to copy value.",
                        copy=dct["OutputValue"],
                        url=url,
                        uid=f"Output {k}",
                    )
                    for k, dct in outputs.items()
                ]
            )

            tags = rl.extract_tags(res)
            detail_items.extend(rl.DetailItem.from_tags(tags, url))
        return detail_items
    # fmt: on


class CloudFormationStackSearcher(rl.BaseSearcher[CloudFormationStack]):
    pass


cloudformation_stack_searcher = CloudFormationStackSearcher(
    # list resources
    service="cloudformation",
    method="list_stacks",
    is_paginator=True,
    default_boto_kwargs={
        "PaginationConfig": {
            "MaxItems": 9999,
        },
    },
    result_path=rl.ResultPath("StackSummaries"),
    # extract document
    doc_class=CloudFormationStack,
    # search
    resource_type=rl.SearcherEnum.cloudformation_stack.value,
    fields=CloudFormationStack.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.cloudformation_stack.value),
    more_cache_key=None,
)



# -*- coding: utf-8 -*-

import zelfred.api as zelfred
from aws_resource_search.items.base_item import BaseArsItem


class TestBaseArsItem:
    def test(self):
        url = "https://www.google.com"
        text = "a1b2c3"
        ui = zelfred.UI(handler=lambda x: x)

        item = BaseArsItem(title="hello")
        # ----------------------------------------------------------------------
        # We only test these two action manually on developer's machine, not in CI.
        # ----------------------------------------------------------------------
        # item.open_url_or_print(ui=ui, url=url)
        # item.copy_or_print(ui=ui, text=text)


if __name__ == "__main__":
    from aws_resource_search.tests.helper import run_cov_test

    run_cov_test(
        __file__,
        "aws_resource_search.items.base_item",
        preview=False,
    )



# -*- coding: utf-8 -*-


from aws_resource_search.conf.define import Config, SearcherEnum
from rich import print as rprint


def test():
    config = Config.load()
    for searcher_enum in SearcherEnum:
        config.get_cache_expire(searcher_enum.value)


if __name__ == "__main__":
    from aws_resource_search.tests.helper import run_cov_test

    run_cov_test(__file__, "aws_resource_search.conf.define", preview=False)



# -*- coding: utf-8 -*-

import typing as T
import dataclasses

import aws_arns.api as arns
import aws_console_url.api as acu

from .. import res_lib as rl

if T.TYPE_CHECKING:
    from ..ars_def import ARS


class IamMixin:
    @property
    def create_date(self: rl.ResourceDocument) -> str:
        return rl.get_datetime_simple_fmt(self.raw_data, "CreateDate")

    @property
    def arn(self: rl.ResourceDocument) -> str:
        return self.raw_data["Arn"]


@dataclasses.dataclass
class IamGroup(IamMixin, rl.ResourceDocument):
    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        return cls(
            raw_data=resource,
            id=resource["GroupName"],
            name=resource["GroupName"],
        )

    @property
    def title(self) -> str:
        return rl.format_key_value("name", self.name)

    @property
    def subtitle(self) -> str:
        return "{}, {}, {}".format(
            rl.format_key_value("create_at", self.create_date),
            rl.format_key_value("arn", self.arn),
            self.short_subtitle,
        )

    @property
    def autocomplete(self) -> str:
        return self.name

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.iam.get_user_group(name_or_arn=self.arn)

    @classmethod
    def get_list_resources_console_url(cls, console: acu.AWSConsole) -> str:
        return console.iam.groups


class IamGroupSearcher(rl.BaseSearcher[IamGroup]):
    pass


iam_group_searcher = IamGroupSearcher(
    # list resources
    service="iam",
    method="list_groups",
    is_paginator=True,
    default_boto_kwargs={"PaginationConfig": {"MaxItems": 9999, "PageSize": 1000}},
    result_path=rl.ResultPath("Groups"),
    # extract document
    doc_class=IamGroup,
    # search
    resource_type=rl.SearcherEnum.iam_group.value,
    fields=IamGroup.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.iam_group.value),
    more_cache_key=None,
)


@dataclasses.dataclass
class IamUser(IamMixin, rl.ResourceDocument):
    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        return cls(
            raw_data=resource,
            id=resource["UserName"],
            name=resource["UserName"],
        )

    @property
    def title(self) -> str:
        return rl.format_key_value("name", self.name)

    @property
    def subtitle(self) -> str:
        return "{}, {}, {}".format(
            rl.format_key_value("create_at", self.create_date),
            rl.format_key_value("arn", self.arn),
            self.short_subtitle,
        )

    @property
    def autocomplete(self) -> str:
        return self.name

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.iam.get_user(name_or_arn=self.arn)

    @classmethod
    def get_list_resources_console_url(cls, console: acu.AWSConsole) -> str:
        return console.iam.users


class IamUserSearcher(rl.BaseSearcher[IamUser]):
    pass


iam_user_searcher = IamUserSearcher(
    # list resources
    service="iam",
    method="list_users",
    is_paginator=True,
    default_boto_kwargs={"PaginationConfig": {"MaxItems": 9999, "PageSize": 1000}},
    result_path=rl.ResultPath("Users"),
    # extract document
    doc_class=IamUser,
    # search
    resource_type=rl.SearcherEnum.iam_user.value,
    fields=IamUser.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.iam_user.value),
    more_cache_key=None,
)


@dataclasses.dataclass
class IamRole(IamMixin, rl.ResourceDocument):
    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        return cls(
            raw_data=resource,
            id=resource["RoleName"],
            name=resource["RoleName"],
        )

    def is_service_role(self) -> bool:
        return "/aws-service-role/" in self.arn

    @property
    def title(self) -> str:
        if self.is_service_role():
            return rl.format_key_value("🪖 name", self.name)
        else:
            return rl.format_key_value("🧢 name", self.name)

    @property
    def subtitle(self) -> str:
        return "{}, {}, {}".format(
            rl.format_key_value("create_at", self.create_date),
            rl.format_key_value("arn", self.arn),
            self.short_subtitle,
        )

    @property
    def autocomplete(self) -> str:
        return self.name

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.iam.get_role(name_or_arn=self.arn)

    @classmethod
    def get_list_resources_console_url(cls, console: acu.AWSConsole) -> str:
        return console.iam.roles

    # fmt: off
    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        from_detail = rl.DetailItem.from_detail
        url = self.get_console_url(console=ars.aws_console)
        detail_items = rl.DetailItem.get_initial_detail_items(doc=self, ars=ars)

        trust_entities = self.raw_data["AssumeRolePolicyDocument"]
        detail_items.extend(
            [
                from_detail("role_id", self.raw_data["RoleId"], url=url),
                from_detail(
                    "trust_entities",
                    trust_entities,
                    self.one_line(trust_entities),
                    url=url,
                ),
            ]
        )

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.iam_client.list_attached_role_policies(
                RoleName=self.name,
                MaxItems=50,
            )
            detail_items.extend([
                from_detail(
                    key="managed policy",
                    value=dct["PolicyArn"],
                    value_text=dct["PolicyName"],
                    url=ars.aws_console.iam.get_policy(name_or_arn=dct["PolicyArn"]),
                )
                for dct in res.get("AttachedPolicies", [])
            ])

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.iam_client.list_role_policies(RoleName=self.name, MaxItems=50)
            detail_items.extend(
                [
                    from_detail(
                        key="inline policy",
                        value=arns.res.IamPolicy.new(
                            aws_account_id=ars.bsm.aws_account_id,
                            name=policy_name,
                        ).to_arn(),
                        value_text=policy_name,
                        url=ars.aws_console.iam.get_role_inline_policy(
                            role_name_or_arn=self.name,
                            policy_name=policy_name,
                        ),
                    )
                    for policy_name in res.get("PolicyNames", [])
                ]
            )

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.iam_client.list_role_tags(RoleName=self.name)
            tags = rl.extract_tags(res)
            detail_items.extend(rl.DetailItem.from_tags(tags, url))

        return detail_items
    # fmt: on


class IamRoleSearcher(rl.BaseSearcher[IamRole]):
    pass


iam_role_searcher = IamRoleSearcher(
    # list resources
    service="iam",
    method="list_roles",
    is_paginator=True,
    default_boto_kwargs={"PaginationConfig": {"MaxItems": 9999, "PageSize": 1000}},
    result_path=rl.ResultPath("Roles"),
    # extract document
    doc_class=IamRole,
    # search
    resource_type=rl.SearcherEnum.iam_role.value,
    fields=IamRole.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.iam_role.value),
    more_cache_key=None,
)


@dataclasses.dataclass
class IamPolicy(IamMixin, rl.ResourceDocument):
    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        return cls(
            raw_data=resource,
            id=resource["PolicyName"],
            name=resource["PolicyName"],
        )

    @property
    def title(self) -> str:
        return rl.format_key_value("name", self.name)

    @property
    def subtitle(self) -> str:
        return "{}, {}, {}".format(
            rl.format_key_value("create_at", self.create_date),
            rl.format_key_value("arn", self.arn),
            self.short_subtitle,
        )

    @property
    def autocomplete(self) -> str:
        return self.name

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.iam.get_policy(name_or_arn=self.arn)

    @classmethod
    def get_list_resources_console_url(cls, console: acu.AWSConsole) -> str:
        return console.iam.policies

    # fmt: off
    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        from_detail = rl.DetailItem.from_detail
        url = self.get_console_url(console=ars.aws_console)
        detail_items = rl.DetailItem.get_initial_detail_items(doc=self, ars=ars)
        detail_items.append(from_detail("create_date", self.create_date))
        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.iam_client.get_policy(PolicyArn=self.arn)
            dct = res["Policy"]
            policy_id = dct["PolicyId"]
            default_version_id = dct["DefaultVersionId"]
            attachment_count = dct["AttachmentCount"]
            description = dct.get("Description", "No description")
            res = ars.bsm.iam_client.get_policy_version(PolicyArn=self.arn, VersionId=default_version_id)
            document: dict = res["PolicyVersion"]["Document"]
            detail_items.extend([
                from_detail("policy_id", policy_id, url=url),
                from_detail("default_version_id", default_version_id, url=url),
                from_detail("attachment_count", attachment_count, url=url),
                from_detail("description", description, url=url),
                from_detail("document", document, self.one_line(document), url=url),
            ])
        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.iam_client.list_policy_tags(PolicyArn=self.arn)
            tags: dict = {dct["Key"]: dct["Value"] for dct in res.get("Tags", [])}
            detail_items.extend(rl.DetailItem.from_tags(tags, url))
        return detail_items
    # fmt: on


class IamPolicySearcher(rl.BaseSearcher[IamPolicy]):
    pass


iam_policy_searcher = IamPolicySearcher(
    # list resources
    service="iam",
    method="list_policies",
    is_paginator=True,
    default_boto_kwargs={"PaginationConfig": {"MaxItems": 9999, "PageSize": 1000}},
    result_path=rl.ResultPath("Policies"),
    # extract document
    doc_class=IamPolicy,
    # search
    resource_type=rl.SearcherEnum.iam_policy.value,
    fields=IamPolicy.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.iam_policy.value),
    more_cache_key=None,
)



# -*- coding: utf-8 -*-

"""
[CN] maintainer note

Resource Search 的核心是根据 searchable document. 一个 Document 代表着一个类似于字典的
数据结构, 只不过它是能被 whoosh 所索引并搜索的. 从 AWS API 上抓下来的数据要被转换成
Document, 保存在 whoosh 的 index 里面. 在搜索的时候, 我们会 match 到一些 Document,
我们将其反序列化为 Document 对象, 然后生成 UI 的 dropdown menu 所需的
:class:`aws_resource_search.items.base_item.BaseArsItem` 对象.`

这个模块定义了所有 Document 的基类以及实体类.
"""



# -*- coding: utf-8 -*-

import typing as T
import dataclasses

import aws_arns.api as arns

import sayt.api as sayt
import aws_arns.api as arns
import aws_console_url.api as acu

from .. import res_lib as rl

if T.TYPE_CHECKING:
    from ..ars_def import ARS


@dataclasses.dataclass
class CodePipelinePipeline(rl.ResourceDocument):
    # fmt: off
    pipeline_arn: str = dataclasses.field(metadata={"field": sayt.StoredField(name="pipeline_arn")})
    # fmt: on

    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        return cls(
            raw_data=resource,
            id=resource["name"],
            name=resource["name"],
            pipeline_arn=arns.res.CodePipelinePipeline.new(
                aws_account_id=bsm.aws_account_id,
                aws_region=bsm.aws_region,
                name=resource["name"],
            ).to_arn(),
        )

    @property
    def title(self) -> str:
        return rl.format_key_value("pipeline_name", self.name)

    @property
    def autocomplete(self) -> str:
        return self.name

    @property
    def arn(self) -> str:
        return self.pipeline_arn

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.codepipeline.get_pipeline(name_or_arn=self.arn)

    @classmethod
    def get_list_resources_console_url(cls, console: acu.AWSConsole) -> str:
        return console.codepipeline.pipelines

    # fmt: off
    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        from_detail = rl.DetailItem.from_detail
        url = self.get_console_url(console=ars.aws_console)
        detail_items = rl.DetailItem.get_initial_detail_items(doc=self, ars=ars)
        
        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.codepipeline_client.get_pipeline(name=self.name)
            dct = res["pipeline"]
            version = dct.get("version", 0)
            update_at = rl.get_none_or_default(dct, "metadata.updated", "NA")
            roleArn = dct.get("roleArn", "NA")
            variables = dct.get("variables", [])
            detail_items.extend([
                from_detail("version", version, url=url),
                from_detail("update_at", update_at, url=url),
                from_detail("roleArn", roleArn, url=ars.aws_console.iam.get_role(roleArn)),
            ])

            for d in variables:
                name = d["name"]
                description = d["description"]
                defaultValue = d["defaultValue"]
                detail_items.append(
                    rl.DetailItem.new(
                        title="🎯 var = {}, default value = {}, ({})".format(
                            rl.format_key(name),
                            rl.format_value(defaultValue),
                            description,
                        ),
                        subtitle=f"🌐 {rl.ShortcutEnum.ENTER} to open url, 📋 {rl.ShortcutEnum.CTRL_A} to copy var name.",
                        uid=f"var {name}",
                        copy=name,
                        url=url,
                    )
                )

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.codepipeline_client.list_tags_for_resource(resourceArn=self.arn)
            tags = rl.extract_tags(res)
            detail_items.extend(rl.DetailItem.from_tags(tags, url))

        return detail_items
    # fmt: on


class CodePipelinePipelineSearcher(rl.BaseSearcher[CodePipelinePipeline]):
    pass


codepipeline_pipeline_searcher = CodePipelinePipelineSearcher(
    # list resources
    service="codepipeline",
    method="list_pipelines",
    is_paginator=True,
    default_boto_kwargs={
        "PaginationConfig": {
            "MaxItems": 5000,
            "PageSize": 1000,
        },
    },
    result_path=rl.ResultPath("pipelines"),
    # extract document
    doc_class=CodePipelinePipeline,
    # search
    resource_type=rl.SearcherEnum.codepipeline_pipeline.value,
    fields=CodePipelinePipeline.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(
        rl.SearcherEnum.codepipeline_pipeline.value
    ),
    more_cache_key=None,
)



# -*- coding: utf-8 -*-

"""
See :class:`InfoItem`.
"""

import dataclasses

from .base_item import BaseArsItem


@dataclasses.dataclass
class InfoItem(BaseArsItem): # pragma: no cover
    """
    Represent an item to show any information. Nothing would happen when user
    press any of the user action key. This is usually used to show some helpful
    information to suggest user what to do.
    """



# -*- coding: utf-8 -*-

import typing as T
import dataclasses
from datetime import datetime

import sayt.api as sayt
import aws_arns.api as arns
import aws_console_url.api as acu

from .. import res_lib as rl

if T.TYPE_CHECKING:
    from ..ars_def import ARS


@dataclasses.dataclass
class SsmParameter(rl.ResourceDocument):
    # fmt: off
    type: str = dataclasses.field(metadata={"field": sayt.NgramWordsField(name="type", minsize=2, maxsize=4, stored=True)})
    tier: str = dataclasses.field(metadata={"field": sayt.NgramWordsField(name="tier", minsize=2, maxsize=4, stored=True)})
    param_arn: str = dataclasses.field(metadata={"field": sayt.StoredField(name="param_arn")})
    # fmt: on

    @property
    def last_modified_date(self) -> datetime:
        return rl.get_datetime(self.raw_data, "LastModifiedDate")

    @property
    def description(self) -> str:
        return rl.get_description(self.raw_data, "Description")

    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        return cls(
            raw_data=resource,
            type=resource["Type"],
            tier=resource["Tier"],
            id=resource["Name"],
            name=resource["Name"],
            param_arn=arns.res.SSMParameter.new(
                aws_account_id=bsm.aws_account_id,
                aws_region=bsm.aws_region,
                name=resource["Name"],
            ).to_arn(),
        )

    @property
    def title(self) -> str:
        return rl.format_key_value("name", self.name)

    @property
    def subtitle(self) -> str:
        return "{}, {}, {}".format(
            rl.format_key_value("type", self.type),
            rl.format_key_value("tier", self.tier),
            self.short_subtitle,
        )

    @property
    def autocomplete(self) -> str:
        return self.name

    @property
    def arn(self) -> str:
        return self.param_arn

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.ssm.get_parameter(name_or_arn=self.arn)

    @classmethod
    def get_list_resources_console_url(cls, console: acu.AWSConsole) -> str:
        return console.ssm.parameters

    # fmt: off
    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        from_detail = rl.DetailItem.from_detail
        url = self.get_console_url(console=ars.aws_console)
        detail_items = rl.DetailItem.get_initial_detail_items(doc=self, ars=ars)

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.ssm_client.get_parameter(Name=self.name)
            param = res["Parameter"]
            detail_items.extend([
                from_detail("LastModifiedDate", self.raw_data.get("LastModifiedDate", "NA"), url=url),
                from_detail("Type", param.get("Type", "NA"), url=url),
                from_detail("Version", param.get("Version", "NA"), url=url),
                from_detail("DataType", param.get("DataType", "NA"), url=url),
                from_detail("Value", param.get("Value", "NA"), self.one_line(param.get("Value", "NA")), url=url),
            ])

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.ssm_client.list_tags_for_resource(
                ResourceType="Parameter",
                ResourceId=self.name,
            )
            tags = rl.extract_tags(res)
            detail_items.extend(rl.DetailItem.from_tags(tags, url))

        return detail_items
    # fmt: on


class SsmParameterSearcher(rl.BaseSearcher[SsmParameter]):
    pass


ssm_parameter_searcher = SsmParameterSearcher(
    # list resources
    service="ssm",
    method="describe_parameters",
    is_paginator=True,
    default_boto_kwargs={"PaginationConfig": {"MaxItems": 9999, "PageSize": 50}},
    result_path=rl.ResultPath("Parameters"),
    # extract document
    doc_class=SsmParameter,
    # search
    resource_type=rl.SearcherEnum.ssm_parameter.value,
    fields=SsmParameter.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.ssm_parameter.value),
    more_cache_key=None,
)



# -*- coding: utf-8 -*-

import typing as T
import random

from .utils import guid, envs, rand_env

if T.TYPE_CHECKING:
    from .main import FakeAws


class VpcMixin:
    vpc_id_list: T.List[str] = list()
    subnet_id_list: T.List[str] = list()
    security_group_id_list: T.List[str] = list()

    @classmethod
    def create_vpc(cls: T.Type["FakeAws"]):
        for ith, env in enumerate(envs, start=1):
            kwargs = {
                "CidrBlock": f"10.{ith}.0.0/16",
                "TagSpecifications": [
                    dict(
                        ResourceType="vpc",
                        Tags=[dict(Key="Name", Value=f"{env}-{guid}-vpc")],
                    )
                ],
            }
            res = cls.bsm.ec2_client.create_vpc(**kwargs)
            vpc_id = res["Vpc"]["VpcId"]
            cls.vpc_id_list.append(vpc_id)

    @classmethod
    def create_subnet(
        cls: T.Type["FakeAws"],
        vpc_id_list: T.Optional[T.List[str]] = None,
    ):
        if vpc_id_list is None:
            vpc_id_list = cls.vpc_id_list
        for ith_vpc, vpc_id in enumerate(vpc_id_list, start=1):
            for ith_subnet, env in enumerate(envs, start=1):
                kwargs = {
                    "CidrBlock": f"10.{ith_vpc}.{ith_subnet}.0/24",
                    "TagSpecifications": [
                        dict(
                            ResourceType="subnet",
                            Tags=[
                                dict(Key="Name", Value=f"{vpc_id}/{env}-{guid}-subnet")
                            ],
                        )
                    ],
                }
                if vpc_id:
                    kwargs["VpcId"] = vpc_id
                res = cls.bsm.ec2_client.create_subnet(**kwargs)
                subnet_id = res["Subnet"]["SubnetId"]
                cls.subnet_id_list.append(subnet_id)

    @classmethod
    def create_security_group(
        cls: T.Type["FakeAws"],
        vpc_id_list: T.Optional[T.List[str]] = None,
    ):
        if vpc_id_list is None:
            vpc_id_list = cls.vpc_id_list
        # create some security group for default vpc and our manually created vpc
        for vpc_id in (None, *vpc_id_list):
            for ith, env in enumerate(envs, start=1):
                kwargs = {
                    "GroupName": f"{env}-{guid}-security-group",
                    "Description": "fake security group",
                    "TagSpecifications": [
                        dict(
                            ResourceType="security-group",
                            Tags=[
                                dict(
                                    Key="Name",
                                    Value=f"{vpc_id}/{env}-{guid}-security-group",
                                )
                            ],
                        )
                    ],
                }
                if vpc_id:
                    kwargs["VpcId"] = vpc_id
                res = cls.bsm.ec2_client.create_security_group(**kwargs)
                sg_id = res["GroupId"]
                cls.security_group_id_list.append(sg_id)



# -*- coding: utf-8 -*-

"""
See :func:`open_config_handler`.
"""

import typing as T

from .. import res_lib as rl
from ..paths import path_config_json

if T.TYPE_CHECKING:  # pragma: no cover
    from ..ui_def import UI


def open_config_handler(
    ui: "UI",
    line_input: str,
    skip_ui: bool = False,
) -> T.List[rl.FileItem]:  # pragma: no cover
    """
    Show current AWS account information of the current boto session.

    :param line_input: the original query input by user before the ``!{``.
    """
    ui.render.prompt = f"(Query)"
    items = [
        rl.FileItem.from_file(
            path=path_config_json,
            title="🛠 Open ${HOME}/.aws_resource_search/config.json",
            uid=f"open-config",
            autocomplete=line_input,
        )
    ]
    return items



# -*- coding: utf-8 -*-

"""
Set up a fake AWS account with a lot of resources using moto for testing.
"""

import moto

from ...ars_def import ARS
from ...ui_def import UI
from ..mock_test import BaseMockTest

from .awslambda import LambdaMixin
from .cloudformation import CloudFormationMixin
from .dynamodb import DynamodbMixin
from .glue import GlueMixin
from .ec2 import Ec2Mixin
from .vpc import VpcMixin
from .s3 import S3Mixin
from .stepfunction import StepFunctionMixin
from .iam import IamMixin


class FakeAws(
    BaseMockTest,
    LambdaMixin,
    CloudFormationMixin,
    DynamodbMixin,
    GlueMixin,
    Ec2Mixin,
    VpcMixin,
    S3Mixin,
    StepFunctionMixin,
    IamMixin,
):
    """
    .. note::

        methods are sorted alphabetically.
    """

    mock_list = [
        moto.mock_apigateway,
        moto.mock_apigatewayv2,
        moto.mock_athena,
        moto.mock_batch,
        moto.mock_cloudformation,
        moto.mock_cloudwatch,
        moto.mock_codebuild,
        moto.mock_codecommit,
        moto.mock_codepipeline,
        moto.mock_dynamodb,
        moto.mock_ec2,
        moto.mock_ecr,
        moto.mock_ecs,
        moto.mock_events,
        moto.mock_firehose,
        moto.mock_glue,
        moto.mock_iam,
        moto.mock_kinesis,
        moto.mock_kms,
        moto.mock_lambda,
        moto.mock_rds,
        moto.mock_s3,
        moto.mock_sagemaker,
        moto.mock_sagemakerruntime,
        moto.mock_secretsmanager,
        moto.mock_ssm,
        moto.mock_stepfunctions,
        moto.mock_sts,
        moto.mock_sqs,
        moto.mock_sns,
    ]

    ars: ARS
    ui: UI

    @classmethod
    def create_all(cls):
        cls.create_cloudformation_stack()
        cls.create_dynamodb_tables()
        cls.create_ec2_instances()
        cls.create_vpc()
        cls.create_subnet()
        cls.create_security_group()
        cls.create_s3_bucket()
        cls.create_iam()
        cls.create_glue_database_table()
        cls.create_glue_job()
        cls.create_glue_crawler()
        cls.create_lambda_layers()
        cls.create_lambda_functions()
        cls.create_state_machines()

    @classmethod
    def setup_ars(cls):
        cls.ars = ARS.from_bsm(bsm=cls.bsm)
        cls.ars.set_profile()
        cls.ars.clear_all_cache()

    @classmethod
    def setup_ui(cls):
        cls.ui = UI.new(ars=cls.ars)




# -*- coding: utf-8 -*-

"""
See :class:`ResourceDocument`.
"""

import typing as T
import uuid
import json
import dataclasses
from datetime import datetime, timezone

import jmespath
import sayt.api as sayt

try:
    import pyperclip
except ImportError:  # pragma: no cover
    pass

from ..terminal import SUBTITLE, SHORT_SUBTITLE
from .base_document import BaseArsDocument

if T.TYPE_CHECKING:  # pragma: no cover
    from boto_session_manager import BotoSesManager
    import aws_console_url.api as acu

    from ..ars_def import ARS
    from ..downloader import T_RESULT_DATA, ResourceIterproxy
    from ..items.api import T_ARS_ITEM


def get_utc_now() -> datetime:
    """
    Get timezone aware datetime object of current time in UTC.
    """
    return datetime.utcnow().replace(tzinfo=timezone.utc)


def to_human_readable_elapsed(sec: int) -> str:
    """
    Convert seconds to human-readable string.

    Example:

        >>> to_human_readable_elapsed(1)
        "1 sec"
        >>> to_human_readable_elapsed(320)
        "5 min 20 sec"
        >>> to_human_readable_elapsed(4200)
        "1 hour 10 min"
        >>> to_human_readable_elapsed(100000)
        "1 day 3.7 hours"
    """
    if sec < 60:
        return f"{sec} sec"
    elif sec < 3600:
        div, mod = divmod(sec, 60)
        if mod:
            return f"{div} min {mod} sec"
        else:
            return f"{div} min"
    elif sec < 86400:
        div, mod = divmod(sec, 3600)
        if mod:
            return f"{div} hour {mod / 60:.1f} min"
        else:
            return f"{div} hour"
    else:
        div, mod = divmod(sec, 86400)
        if mod:
            return f"{div} day {mod / 3600:.1f} hour"
        else:
            return f"{div} day"


def to_utc_dt(dt: datetime) -> datetime:
    """
    Convert to timezone aware datetime object in UTC.

    If the input datetime object doesn't have timezone, it will assume it's in UTC.
    """
    if dt.tzinfo is None:
        return dt.replace(tzinfo=timezone.utc)
    else:
        return dt.astimezone(timezone.utc)


SIMPLE_DT_FMT = "%Y-%m-%d %H:%M:%S"


def to_simple_dt_fmt(dt: datetime) -> str:
    """
    To simple datetime format string.
    """
    return dt.strftime(SIMPLE_DT_FMT)


def to_iso_dt_fmt(dt: datetime) -> str:
    """
    To ISO datetime format string.
    """
    return dt.isoformat()


def get_none_or_default(
    data: "T_RESULT_DATA",
    path: str,
    default: T.Optional[T.Any] = None,
) -> T.Optional[T.Any]:
    """
    A helper method to get value from ``data``. If the value doesn't exist,
    it will return the default value.

    Example:

        >>> doc = {"a": 1, "b": {"c": 3})
        >>> get_none_or_default(data, "a")
        1
        >>> get_none_or_default(data, "d")
        None
        >>> get_none_or_default(data, "d", "hello")
        "hello"
        >>> get_none_or_default(data, "b.c")
        3
        >>> get_none_or_default(data, "b.d")
        None
        >>> get_none_or_default(data, "b.d", "hello")
        "hello"
        >>> get_none_or_default(data, "c.e")
        None
        >>> get_none_or_default(data, "c.e", "hello")
        "hello"

    :param data: The data to get value from. It represent an AWS resource
        in the boto3 API response. Usually it can be a dictionary or a string.
    :param path: jmespath syntax
    :param default: default value if the value doesn't exist
    """
    value = jmespath.search(path, data)
    return default if value is None else value


def get_description(
    data: "T_RESULT_DATA",
    path: str,
    default: T.Optional[str] = "No description",
) -> str:
    desc = get_none_or_default(data, path)
    if desc:
        return desc
    else:
        return default


def get_datetime(
    data: "T_RESULT_DATA",
    path: str,
    default: T.Optional[datetime] = datetime(1970, 1, 1, tzinfo=timezone.utc),
    as_utc: bool = True,
) -> T.Optional[datetime]:
    """
    Extract timezone aware datetime object from ``data`` using Jmespath.

    Example:

        >>> get_datetime({"CreateDate": datetime(2021, 1, 1)}, path="CreateDate")
        datetime(2021, 1, 1)
    """
    dt = get_none_or_default(data, path, default)
    if dt is None:
        return dt

    if as_utc:
        return to_utc_dt(dt)
    else:
        return dt


def get_datetime_simple_fmt(
    data: "T_RESULT_DATA",
    path: str,
    default: str = "No datetime",
    as_utc: bool = True,
) -> str:
    """
    Extract isoformat datetime string from a dictionary using Jmespath.

    Example:

        >>> get_datetime_simple_fmt({"CreateDate": datetime(2021, 1, 1, microsecond=123000)}, path="CreateDate")
        "2021-01-01 00:00:00"
    """
    dt = get_datetime(data, path, default=None, as_utc=as_utc)
    if dt is None:
        return default
    else:
        return to_simple_dt_fmt(dt)


def get_datetime_iso_fmt(
    data: "T_RESULT_DATA",
    path: str,
    default: str = "No datetime",
    as_utc: bool = True,
) -> str:
    """
    Extract isoformat datetime string from a dictionary using Jmespath.

    Example:

        >>> get_datetime_iso_fmt({"CreateDate": datetime(2021, 1, 1)}, path="CreateDate")
        "2021-01-01T00:00:00"
    """
    dt = get_datetime(data, path, default=None, as_utc=as_utc)
    if dt is None:
        return default
    else:
        return to_iso_dt_fmt(dt)


@dataclasses.dataclass
class ResourceDocument(BaseArsDocument):
    """
    This is the base class for searchable AWS resource document for all
    AWS resource type. Also, it provides some utility functions to help with
    writing the concrete document class for a specific AWS resource type.

    A 'document' is a searchable object stored in the index, representing
    the detailed data of an AWS resource.

    In the ``aws_resource_search.res.${service}.py`` module, we need to inherit
    this class to create a document class for each AWS resource type. For example,
    you can define a document class for S3 bucket like this:

        >>> @dataclasses.dataclass
        >>> class S3BucketDocument(ResourceDocument):
        ...    pass

    :param raw_data: The raw data from the boto3 API call response.
        For example, for
        `s3_client.list_buckets <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3/client/list_buckets.html>`,
        the raw data is
        ``{'Name': 'string', 'CreationDate': datetime(2015, 1, 1)}``
    :param id: the unique id of the aws resource, most of the time it's the
        same as the resource name. For example, AWS Lambda function name,
        S3 bucket name. But for some resources, it's not. For example,
        AWS EC2 instance id, AWS VPC id. This field is also used as a searchable
        ``IdField`` in :class:`aws_resource_search.base_searcher.BaseSearcher`.
    :param name: the human friendly name of the aws resource. For example,
        AWS Lambda function name, S3 bucket name. This field is also used as a
        searchable ``NgramWordsField`` in :class:`aws_resource_search.base_searcher.BaseSearcher`.
    :param name: similar to name, but this field is used as a searchable
        ``TextField`` in :class:`aws_resource_search.base_searcher.BaseSearcher`.

    In your subclass, you must implement the following methods.
    Please read the docstrings to understand their functionality.

    - :meth:`from_resource`
    - :meth:`title`
    - :meth:`arn`
    - :meth:`get_console_url`

    **Subclass FAQ**

    - Q: I see there's an attribute in the raw data from the boto3 API response,
        should I create a ``dataclasses.field()`` for it?
    - A: Depends. If you want it to be searchable, yes, declare a field for it.
        If you don't want it to be searchable, and you just want to reference
        it later, declare a ``@property`` attribute instead.

    - Q: I have a derivable attribute can be calculated from the raw data,
        should I create a ``dataclasses.field()`` for it?
    - A: Depends.
        1. if you want it to be searchable, yes, declare a field for it.
        2. if the value can be calculated only using the raw data,
            declare a ``@property`` attribute instead.
        3. if the calculation needs additional information like the
            ``BotoSesManager`` object you used for the boto3 API call,
            then you need to create a ``dataclasses.field()`` for it.
            Because you can't calculate it again when getting it back
            from the search index. At that time, you don't have access
            to the original ``BotoSesManager`` object.
    """

    # fmt: off
    raw_data: "T_RESULT_DATA" = dataclasses.field(metadata={"field": sayt.StoredField(name="raw_data")})
    id: str = dataclasses.field(metadata={"field": sayt.IdField(name="id", field_boost=5.0, stored=True)})
    name: str = dataclasses.field(metadata={"field": sayt.NgramWordsField(name="name", minsize=2, maxsize=4, stored=True, sortable=True, ascending=True)})
    name_text: str = dataclasses.field(metadata={
        "field": sayt.TextField(name="name_text", stored=False, sortable=True, ascending=True)}, init=False)
    # fmt: on

    def __post_init__(self):
        name_text = self.name
        for char in "-_":
            name_text = name_text.replace(char, " ")
        self.name_text = name_text

    @classmethod
    def from_resource(
        cls,
        resource: "T_RESULT_DATA",
        bsm: "BotoSesManager",
        boto_kwargs: dict,
    ):
        """
        Create a document object from the boto3 API response data.
        Since the API response data relies on both the 'bsm' object
        and 'boto_kwargs,' we also require this information to
        create the document object.

        For example, ``s3_client.list_buckets`` api returns::

            {
                'Name': '...',
                'CreationDate': datetime.datetime(...)
            }

        Then, the implementation should be::

            >>> @dataclasses.dataclass
            ... class S3BucketDocument(ResourceDocument):
            ...     @classmethod
            ...     def from_resource(
            ...         cls,
            ...         resource: "T_RESULT_DATA",
            ...         bsm: "BotoSesManager",
            ...         boto_kwargs: dict,
            ...     ):
            ...         return cls(
            ...             raw_data=resource,
            ...             id=resource["Name"],
            ...             name=resource["Name"],
            ...         )

        .. note::

            You have to implement this method for each AWS Resource Type.
        """
        raise NotImplementedError

    @classmethod
    def from_many_resources(
        cls,
        resources: "ResourceIterproxy",
        bsm: "BotoSesManager",
        boto_kwargs: dict,
    ):
        """
        A wrapper of :meth:`~ResourceDocument.from_resource` to create a list of
        document objects.
        """
        for resource in resources:
            yield cls.from_resource(
                resource=resource,
                bsm=bsm,
                boto_kwargs=boto_kwargs,
            )

    @property
    def title(self) -> str:
        """
        The title in the zelfred UI.

        For example, you want to use the S3 bucket name as the title::

            >>> @dataclasses.dataclass
            ... class S3BucketDocument(ResourceDocument):
            ...     @property
            ...     def title(self) -> str:
            ...         return f"bucket name = {self.name}"

        .. important::

            You have to implement this method for each AWS Resource Type.
        """
        raise NotImplementedError

    @property
    def subtitle(self) -> str:
        """
        The subtitle in the zelfred UI.

        .. seealso::

            :const:`~aws_resource_search.terminal.SUBTITLE`
        """
        return SUBTITLE

    @property
    def short_subtitle(self) -> str:
        """
        A shorter version of subtitle.

        .. seealso::

            :const:`~aws_resource_search.terminal.SHORT_SUBTITLE`
        """
        return SHORT_SUBTITLE

    @property
    def uid(self) -> str:
        """
        The internal uid used for sorting and deduplication in the zelfred UI.
        """
        return uuid.uuid4().hex

    @property
    def autocomplete(self) -> str:
        """
        Autocomplete text for the zelfred UI.
        """
        return ""

    @property
    def arn(self) -> str:
        """
        AWS Resource ARN, if applicable. User can tap 'Ctrl + A' to copy the ARN.

        If the ``raw_data`` does not sufficient information to construct the ARN,
        you should declare a ``dataclasses.field()`` for it and initialize it
        in the :meth:`~ResourceDocument.from_resource` method. Then you can just
        reference it here.

        .. important::

            You have to implement this method for each AWS Resource Type.
        """
        msg = f"{self.__class__.__name__} doesn't support ARN"
        raise NotImplementedError(msg)

    def get_console_url(self, console: "acu.AWSConsole") -> str:
        """
        AWS Console URL to view this AWS resource in the console.
        If applicable, User can tap 'Enter' to open in browser.

        .. important::

            You have to implement this method for each AWS Resource Type.
        """
        msg = f"{self.__class__.__name__} doesn't support AWS Console url"
        raise NotImplementedError(msg)

    @classmethod
    def get_list_resources_console_url(cls, console: "acu.AWSConsole") -> str:
        """
        AWS Console URL to view list all AWS resources of this type in the console.
        If applicable, User can tap 'Enter' to open in browser.

        .. important::

            You have to implement this method for each AWS Resource Type.
        """
        msg = f"{cls.__name__} doesn't support list resources AWS Console url"
        raise NotImplementedError(msg)

    @staticmethod
    def one_line(obj, na: str = "NA") -> str:
        """
        Convert a python object to one line json string.

        This is usually used in the :meth:`ResourceDocument.get_details`` method
        to normalize multiline json object in the UI.
        """
        if obj:
            if isinstance(obj, str):
                if obj.startswith("[") or obj.startswith("{"):
                    return json.dumps(json.loads(obj))
                else:
                    lines = obj.splitlines()
                    if len(lines) > 1:
                        return f"{lines[0]} ..."
                    else:
                        return lines[0]
            else:
                return json.dumps(obj)
        else:
            return na

    def get_details(self, ars: "ARS") -> T.List["T_ARS_ITEM"]:
        """
        Call boto3 API to get additional details for this resource. The detailed
        information will be rendered in the dropdown menu. It allows user to tap
        'Ctrl + P' to enter a sub-session to view the details. User can tap
        'F1' to go back to the previous session.

        .. important::

            You have to implement this method for each AWS Resource Type.

        .. note::

            Here's some tips to implement this method correctly:

            1. For immutable attributes, such as resource id, name, you can
                extract them from the ``raw_data``. For example, the Ec2 instance id.
            2. For mutable attributes, such as tags, you have to call boto3 API,
                because the ``raw_data`` is from the index and could be outdated.
            3. Use :meth:`~ResourceDocument.enrich_details` context manager when you
                need to call boto3 API. Because it may fail due to permission issue,
                or resource not exists issue.
        """
        msg = f"{self.__class__.__name__} doesn't support get details"
        raise NotImplementedError(msg)

    @classmethod
    def get_dataset_fields(
        cls,
        id_field_boost: float = 5.0,
        name_minsize: int = 2,
        name_maxsize: int = 4,
        name_sortable: bool = True,
        name_ascending: bool = True,
    ) -> T.List[sayt.T_Field]:
        """
        A helper function to generate the arguments for
        :attr:`Searcher.fields` property.

        By default, it will generate three fields:

        - ``raw_data`` (``StoredField``), to store the raw data from boto3 API response.
        - ``id`` (``IdField``), this field should have higher weight (in id_field_boost).
        - ``name`` (``NgramWordsField``), the default setting is that the result
            is ordered by name in ascending order (in name_sortable and name_ascending).

        Also, for each ``dataclasses.field`` you declared, it will use the
        search field declaration stored in
        ``dataclasses.field(..., metadata={"field": sayt.SomeField(name=...)})``.
        This method performs basic validation on the field declaration.
        This method helps your AWS resource document data model and search index
        schema in sync and avoid human-mistake.

        :param id_field_boost: parameter for the id field.
        :param name_minsize: parameter for the name fields.
        :param name_maxsize: parameter for the name fields.
        :param name_sortable: parameter for the name fields.
        :param name_ascending: parameter for the name fields.
        """
        # search_fields = [
        #     sayt.StoredField(name="raw_data"),
        #     sayt.IdField(name="id", field_boost=id_field_boost, stored=True),
        #     sayt.NgramWordsField(
        #         name="name",
        #         minsize=name_minsize,
        #         maxsize=name_maxsize,
        #         stored=True,
        #         sortable=name_sortable,
        #         ascending=name_ascending,
        #     ),
        # ]
        # _set = {"raw_data", "id", "name"}
        search_field_mapper = {}
        for field in dataclasses.fields(cls):
            try:
                search_field = field.metadata["field"]
            except KeyError:
                raise KeyError(
                    "you forget to define "
                    '``dataclasses.field(..., metadata={"field": sayt.SomeField(name="...")})`` '
                    f"for ``{cls.__name__}.{field.name}`` field"
                )

            if not isinstance(search_field, sayt.BaseField):
                raise TypeError(
                    f"the search field in ``{cls.__name__}.{field.name}`` "
                    f"is not a ``sayt.BaseField`` instance"
                )

            if search_field.name != field.name:
                raise ValueError(
                    f"dataclass field name {field.name!r} doesn't match "
                    f"search field name {search_field.name!r}"
                )
            search_field_mapper[search_field.name] = search_field

        search_field_mapper["id"] = dataclasses.replace(
            search_field_mapper["id"],
            field_boost=id_field_boost,
        )
        search_field_mapper["name"] = dataclasses.replace(
            search_field_mapper["name"],
            minsize=name_minsize,
            maxsize=name_maxsize,
            sortable=name_sortable,
            ascending=name_ascending,
        )
        return list(search_field_mapper.values())


T_ARS_RESOURCE_DOCUMENT = T.TypeVar("T_ARS_RESOURCE_DOCUMENT", bound=ResourceDocument)



# -*- coding: utf-8 -*-

import typing as T
import dataclasses

import sayt.api as sayt
import aws_console_url.api as acu

from .. import res_lib as rl

if T.TYPE_CHECKING:
    from ..ars_def import ARS


@dataclasses.dataclass
class RdsDbInstance(rl.ResourceDocument):
    # fmt: off
    status: str = dataclasses.field(metadata={"field": sayt.NgramWordsField(name="status", minsize=2, maxsize=4, stored=True)})
    engine: str = dataclasses.field(metadata={"field": sayt.NgramWordsField(name="engine", minsize=2, maxsize=4, stored=True)})
    # fmt: on

    @property
    def klass(self) -> str:
        return self.raw_data.get("DBInstanceClass", "Unknown")

    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        return cls(
            raw_data=resource,
            status=resource["DBInstanceStatus"],
            engine=resource.get("Engine", "Unknown"),
            id=resource["DBInstanceIdentifier"],
            name=resource["DBInstanceIdentifier"],
        )

    @property
    def title(self) -> str:
        return rl.format_key_value("identifier", self.name)

    @property
    def subtitle(self) -> str:
        return "{}, {}, {}, {}".format(
            rl.format_key_value("status", self.status),
            rl.format_key_value("class", self.klass),
            rl.format_key_value("engine", self.engine),
            self.short_subtitle,
        )

    @property
    def autocomplete(self) -> str:
        return self.name

    @property
    def arn(self) -> str:
        return self.raw_data["DBInstanceArn"]

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.rds.get_database_instance(id_or_arn=self.arn)

    @classmethod
    def get_list_resources_console_url(cls, console: acu.AWSConsole) -> str:
        return console.rds.databases

    # fmt: off
    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        from_detail = rl.DetailItem.from_detail
        url = self.get_console_url(console=ars.aws_console)
        detail_items = rl.DetailItem.get_initial_detail_items(doc=self, ars=ars)

        detail_items.append(
            from_detail("DBInstanceIdentifier", self.id, url=url),
        )

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.rds_client.describe_db_instances(DBInstanceIdentifier=self.name)
            dbs = res.get("DBInstances", [])
            if len(dbs) == 0:
                return [
                    rl.DetailItem.new(
                        title="🚨 DB instance not found, maybe it's deleted?",
                        subtitle=f"{rl.ShortcutEnum.ENTER} to verify in AWS Console",
                        url=url,
                    )
                ]
            db = dbs[0]
            DBInstanceStatus = db["DBInstanceStatus"]
            DBInstanceClass = db.get("DBInstanceClass", "Unknown")
            Engine = db.get("Engine", "Unknown")
            EngineVersion = db.get("EngineVersion", "Unknown")
            Host = db.get("Endpoint", {}).get("Address", "Unknown")
            Port = db.get("Endpoint", {}).get("Port", "Unknown")
            MasterUsername = db.get("Engine", "Unknown")
            DBName = db.get("Engine", "Unknown")
            MultiAZ = db.get("MultiAZ", "Unknown")
            AvailabilityZone = db.get("AvailabilityZone", "Unknown")
            AutoMinorVersionUpgrade = db.get("AutoMinorVersionUpgrade", "Unknown")
            PubliclyAccessible = db.get("PubliclyAccessible", "Unknown")
            StorageType = db.get("StorageType", "Unknown")
            AllocatedStorage = db.get("AllocatedStorage", "Unknown")
            StorageEncrypted = db.get("StorageEncrypted", "Unknown")
            DeletionProtection = db.get("DeletionProtection", "Unknown")
            DBClusterIdentifier = db.get("DBClusterIdentifier", "Unknown")

            detail_items.extend([
                from_detail("DBInstanceStatus", DBInstanceStatus, url=url),
                from_detail("DBInstanceClass", DBInstanceClass, url=url),
                from_detail("Engine", Engine, url=url),
                from_detail("EngineVersion", EngineVersion, url=url),
                from_detail("Host", Host, url=url),
                from_detail("Port", Port, url=url),
                from_detail("MasterUsername", MasterUsername, url=url),
                from_detail("DBName", DBName, url=url),
                from_detail("MultiAZ", MultiAZ, url=url),
                from_detail("AvailabilityZone", AvailabilityZone, url=url),
                from_detail("AutoMinorVersionUpgrade", AutoMinorVersionUpgrade, url=url),
                from_detail("PubliclyAccessible", PubliclyAccessible, url=url),
                from_detail("StorageType", StorageType, url=url),
                from_detail("AllocatedStorage", AllocatedStorage, url=url),
                from_detail("StorageEncrypted", StorageEncrypted, url=url),
                from_detail("DeletionProtection", DeletionProtection, url=url),
                from_detail("DBClusterIdentifier", DBClusterIdentifier, url=ars.aws_console.rds.get_database_cluster(id_or_arn=DBClusterIdentifier)),
            ])

            tags = rl.extract_tags(res)
            detail_items.extend(rl.DetailItem.from_tags(tags, url))
        return detail_items
    # fmt: on


class RdsDbInstanceSearcher(rl.BaseSearcher[RdsDbInstance]):
    pass


rds_db_instance_searcher = RdsDbInstanceSearcher(
    # list resources
    service="rds",
    method="describe_db_instances",
    is_paginator=True,
    default_boto_kwargs={
        "PaginationConfig": {
            "MaxItems": 9999,
            "PageSize": 100,
        },
    },
    result_path=rl.ResultPath("DBInstances"),
    # extract document
    doc_class=RdsDbInstance,
    # search
    resource_type=rl.SearcherEnum.rds_db_instance.value,
    fields=RdsDbInstance.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.rds_db_instance.value),
    more_cache_key=None,
)


@dataclasses.dataclass
class RdsDbCluster(rl.ResourceDocument):
    # fmt: off
    status: str = dataclasses.field(metadata={"field": sayt.NgramWordsField(name="status", minsize=2, maxsize=4, stored=True)})
    engine: str = dataclasses.field(metadata={"field": sayt.NgramWordsField(name="engine", minsize=2, maxsize=4, stored=True)})
    # fmt: on

    @property
    def klass(self) -> str:
        return self.raw_data.get("DBInstanceClass", "Unknown")

    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        return cls(
            raw_data=resource,
            status=resource["Status"],
            engine=resource.get("Engine", "Unknown"),
            id=resource["DBClusterIdentifier"],
            name=resource["DBClusterIdentifier"],
        )

    @property
    def title(self) -> str:
        return rl.format_key_value("identifier", self.name)

    @property
    def subtitle(self) -> str:
        return "{}, {}, {}, {}".format(
            rl.format_key_value("status", self.status),
            rl.format_key_value("class", self.klass),
            rl.format_key_value("engine", self.engine),
            self.short_subtitle,
        )

    @property
    def autocomplete(self) -> str:
        return self.name

    @property
    def arn(self) -> str:
        return self.raw_data["DBClusterArn"]

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.rds.get_database_cluster(id_or_arn=self.arn)

    @classmethod
    def get_list_resources_console_url(cls, console: acu.AWSConsole) -> str:
        return console.rds.databases

    # fmt: off
    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        from_detail = rl.DetailItem.from_detail
        url = self.get_console_url(console=ars.aws_console)
        detail_items = rl.DetailItem.get_initial_detail_items(doc=self, ars=ars)

        detail_items.append(
            from_detail("DBClusterIdentifier", self.id, url=url),
        )
        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.rds_client.describe_db_clusters(DBClusterIdentifier=self.name)
            dbs = res.get("DBClusters", [])
            if len(dbs) == 0:
                return [
                    rl.DetailItem.new(
                        title="🚨 DB cluster not found, maybe it's deleted?",
                        subtitle=f"{rl.ShortcutEnum.ENTER} to verify in AWS Console",
                        url=url,
                    )
                ]
            db = dbs[0]

            Status = db["Status"]
            engine = db.get("Engine", "Unknown")
            EngineVersion = db.get("EngineVersion", "Unknown")
            Endpoint = db.get("Endpoint", "Unknown")
            ReaderEndpoint = db.get("ReaderEndpoint", "Unknown")
            CustomEndpoints = db.get("CustomEndpoints", "Unknown")
            Port = db.get("Port", "Unknown")
            MasterUsername = db.get("Engine", "Unknown")
            DatabaseName = db.get("DatabaseName", "Unknown")
            MultiAZ = db.get("MultiAZ", "Unknown")
            AvailabilityZones = db.get("AvailabilityZones", "Unknown")
            IAMDatabaseAuthenticationEnabled = db.get("IAMDatabaseAuthenticationEnabled", "Unknown")
            DeletionProtection = db.get("DeletionProtection", "Unknown")
            DBClusterMembers = db.get("DBClusterMembers", [])
            DBClusterInstanceClass = db.get("DBClusterInstanceClass", "Unknown")

            detail_items.extend([
                from_detail("Status", Status, url=url),
                from_detail("engine", engine, url=url),
                from_detail("EngineVersion", EngineVersion, url=url),
                from_detail("Endpoint", Endpoint, url=url),
                from_detail("ReaderEndpoint", ReaderEndpoint, url=url),
                from_detail("CustomEndpoints", CustomEndpoints, url=url),
                from_detail("Port", Port, url=url),
                from_detail("MasterUsername", MasterUsername, url=url),
                from_detail("DatabaseName", DatabaseName, url=url),
                from_detail("MultiAZ", MultiAZ, url=url),
                from_detail("AvailabilityZones", AvailabilityZones, url=url),
                from_detail("IAMDatabaseAuthenticationEnabled", IAMDatabaseAuthenticationEnabled, url=url),
                from_detail("DeletionProtection", DeletionProtection, url=url),
                *[
                    from_detail(
                        f"🖥️ db instance",
                        member["DBInstanceIdentifier"],
                        url=ars.aws_console.rds.get_database_instance(id_or_arn=member["DBInstanceIdentifier"]),
                        uid="db instance {}".format(member["DBInstanceIdentifier"])
                    )
                    for member in DBClusterMembers
                ],
                from_detail("DBClusterInstanceClass", DBClusterInstanceClass, url=url),
            ])

            tags = rl.extract_tags(res)
            detail_items.extend(rl.DetailItem.from_tags(tags, url))
        return detail_items
    # fmt: on


class RdsDbClusterSearcher(rl.BaseSearcher[RdsDbCluster]):
    pass


rds_db_cluster_searcher = rl.BaseSearcher(
    # list resources
    service="rds",
    method="describe_db_clusters",
    is_paginator=True,
    default_boto_kwargs={
        "PaginationConfig": {
            "MaxItems": 9999,
        },
    },
    result_path=rl.ResultPath("DBClusters"),
    # extract document
    doc_class=RdsDbCluster,
    # search
    resource_type=rl.SearcherEnum.rds_db_cluster.value,
    fields=RdsDbCluster.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.rds_db_cluster.value),
    more_cache_key=None,
)



# -*- coding: utf-8 -*-

import dataclasses

import pytest

import json
from datetime import datetime, timezone

import botocore.exceptions

import sayt.api as sayt
from aws_resource_search.documents.resource_type_document import (
    ResourceTypeDocument,
)


class TestResourceTypeDocument:
    def test_get_dataset_fields(self):
        ResourceTypeDocument.get_dataset_fields()


if __name__ == "__main__":
    from aws_resource_search.tests.helper import run_cov_test

    run_cov_test(
        __file__,
        "aws_resource_search.documents.resource_type_document",
        preview=False,
    )



# -*- coding: utf-8 -*-

"""
See :class:`ARSBase`.
"""

import typing as T
import shutil
import dataclasses
from pathlib import Path

from diskcache import Cache
from boto_session_manager import BotoSesManager
from boto_session_manager.manager import NOTHING
import aws_console_url.api as acu

from .exc import MalformedBotoSessionError
from .paths import dir_index, dir_cache
from .searcher_finder import SearcherFinder, searcher_finder
from .ars_search_patterns import ArsSearchPatternsMixin
from .ars_mixin import ARSMixin

if T.TYPE_CHECKING:  # pragma: no cover
    from .res_lib import T_SEARCHER


def validate_bsm(bsm: "BotoSesManager"):
    """
    validate the boto session manager. It has to have ``sts.get_caller_identity()``
    permission because we need to get aws_account_id, and be able to get ``aws_region``.
    """
    try:
        _ = bsm.aws_account_id
    except Exception as e:  # pragma: no cover
        raise MalformedBotoSessionError(
            f"failed to use sts.get_caller_identity() to get AWS account id:, "
            f"please check your default AWS profile in ~/.aws/config and ~/.aws/credentials, "
            f"error: {e}"
        )

    try:
        _ = bsm.aws_region
    except Exception as e:  # pragma: no cover
        raise MalformedBotoSessionError(
            f"failed to get AWS region of your boto session, "
            f"please check your default AWS profile in ~/.aws/config and ~/.aws/credentials, "
            f"error: {e}"
        )


@dataclasses.dataclass
class ARS(
    ArsSearchPatternsMixin,
    ARSMixin,
):
    """
    This class stands for "AWS Resource Search Base", provides method
    to search AWS resources in Python.

    In order to enable auto-complete in IDE, we need to explicitly declare
    per-resource-type searcher method in this class. However, we don't want
    to maintain this wiring logics. Instead, we use Jinja template to generate,
    a code

    Usage::

        # import the singleton ARS object
        >>> from aws_resource_search.api import ars
        >>> ars.s3_bucket.search("my bucket")
        [
            {
                "id": "my-bucket",
                "name": "my-bucket",
                ...
            },
            {
                ...
            },
        ]

    This class is a singleton object that holds all context data such as
    ``boto_session_manager.BotoSesManager``, ``aws_console_url.api.AwsConsole``.
    """
    # fmt: off
    bsm: "BotoSesManager" = dataclasses.field()
    aws_console: "aws_console_url.AWSConsole" = dataclasses.field()
    searcher_finder: "SearcherFinder" = dataclasses.field(default_factory=lambda: searcher_finder)
    dir_index: Path = dataclasses.field(default=dir_index)
    dir_cache: Path = dataclasses.field(default=dir_cache)
    cache: Cache = dataclasses.field(default=None)
    # fmt: on

    def __post_init__(self):
        self.dir_index = Path(self.dir_index)
        if self.dir_cache is not None:  # pragma: no cover
            self.dir_cache = Path(self.dir_cache)
        if self.cache is None:
            self.cache = Cache(str(self.dir_cache), disk_pickle_protocol=5)
        else:  # pragma: no cover
            self.dir_cache = Path(self.cache.directory)

    @classmethod
    def from_profile(cls, profile: T.Optional[str] = NOTHING):
        """
        Create a new :class:`ARSBase` object by an AWS profile. If None, then
        use the default AWS profile.
        """
        bsm = BotoSesManager(profile_name=profile)
        return cls.from_bsm(bsm=bsm)

    @classmethod
    def from_bsm(cls, bsm: T.Optional["BotoSesManager"] = None):
        if bsm is None:
            bsm = BotoSesManager()
        return cls(bsm=bsm, aws_console=acu.AWSConsole.from_bsm(bsm))

    def get_searcher(self, resource_type: str) -> "T_SEARCHER":
        """
        Get corresponding :class:`aws_resource_search.res_lib.Searcher`
        object by resource type.
        """
        sr = searcher_finder.import_searcher(resource_type)
        sr.bsm = self.bsm
        return sr

    def clear_all_cache(self):
        """
        Clear all cache.
        """
        shutil.rmtree(self.dir_index, ignore_errors=True)
        shutil.rmtree(self.dir_cache, ignore_errors=True)

    def all_resource_types(self) -> T.List[str]:
        """
        Return all resource types.
        """
        return searcher_finder.all_resource_types()

    def is_valid_resource_type(self, resource_type: str) -> bool:
        """
        Check if the resource type is supported.
        """
        return searcher_finder.is_valid_resource_type(resource_type)

    def set_profile(self, profile: T.Optional[str] = NOTHING):
        """
        Set all boto session related attributes (``bsm``, ``aws_console``,
        ``search_patterns``) to use a new AWS profile.

        Logics:

        1. Update the singleton ``bsm`` object by reset its cache and use a new AWS profile.
        2. Update the ``aws_console`` object to use the new ``bsm``
        3. Since we updated the ``bsm``, we also need to reset the ``searcher_finder``
            cache to recreate all ``Searcher`` objects.
        """
        # reset boto_session_manager.BotoSesManager
        self.bsm.aws_access_key_id = NOTHING
        self.bsm.aws_secret_access_key = NOTHING
        self.bsm.aws_session_token = NOTHING
        self.bsm.region_name = NOTHING
        self.bsm.botocore_session = NOTHING
        self.bsm.profile_name = profile
        self.bsm.clear_cache()
        validate_bsm(self.bsm)

        # reset aws_console_url.AWSConsole
        self.aws_console = acu.AWSConsole.from_bsm(self.bsm)

        # reset searcher_finder.SearcherFinder
        self.searcher_finder.searcher_cache.clear()

        # reset ARS.search_patterns
        _ = self.search_patterns
        self._clear_search_patterns_cache()



# -*- coding: utf-8 -*-

"""
Usage::

    from .utils import guid, envs, rand_env
"""

import random
import string
from faker import Faker

fake = Faker()

_charset = string.ascii_lowercase + string.digits
guid = "".join([random.choice(_charset) for i in range(4)])
envs = ["sbx", "tst", "prd"]


def rand_env() -> str:
    return random.choice(envs)



# -*- coding: utf-8 -*-

import typing as T
import dataclasses
from datetime import datetime

import sayt.api as sayt
import aws_arns.api as arns
import aws_console_url.api as acu

from .. import res_lib as rl

if T.TYPE_CHECKING:
    from ..ars_def import ARS


def make_env_var_item(
    name: str,
    value: str,
    type: str,
    url: str,
) -> rl.DetailItem:
    return rl.DetailItem.new(
        title="🎯 env var: {} ({})".format(
            rl.format_key_value(name, value),
            type,
        ),
        subtitle=f"🌐 {rl.ShortcutEnum.ENTER} to open url, 📋 {rl.ShortcutEnum.CTRL_A} to copy value.",
        uid=f"env var {name}",
        copy=value,
        url=url,
    )


@dataclasses.dataclass
class CodeBuildProject(rl.ResourceDocument):
    # fmt: off
    project_arn: str = dataclasses.field(metadata={"field": sayt.StoredField(name="project_arn")})
    # fmt: on

    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        return cls(
            raw_data=resource,
            id=resource,
            name=resource,
            project_arn=arns.res.CodeBuildProject.new(
                aws_account_id=bsm.aws_account_id,
                aws_region=bsm.aws_region,
                name=resource,
            ).to_arn(),
        )

    @property
    def title(self) -> str:
        return rl.format_key_value("project_name", self.name)

    @property
    def autocomplete(self) -> str:
        return self.name

    @property
    def arn(self) -> str:
        return self.project_arn

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.codebuild.get_project(project_or_arn=self.arn)

    @classmethod
    def get_list_resources_console_url(cls, console: acu.AWSConsole) -> str:
        return console.codebuild.build_projects

    # fmt: off
    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        from_detail = rl.DetailItem.from_detail
        url = self.get_console_url(console=ars.aws_console)
        detail_items = rl.DetailItem.get_initial_detail_items(doc=self, ars=ars)

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.codebuild_client.batch_get_projects(names=[self.name])
            projects = res.get("projects", [])
            if len(projects) == 0:
                return [
                    rl.DetailItem.new(
                        title="🚨 Project not found, maybe it's deleted?",
                        subtitle=f"{rl.ShortcutEnum.ENTER} to verify in AWS Console",
                        url=self.get_console_url(ars.aws_console),
                    )
                ]
            dct = projects[0]
            description = dct.get("description", "NA")
            serviceRole = dct.get("serviceRole", "NA")
            concurrentBuildLimit = dct.get("concurrentBuildLimit", "NA")
            timeoutInMinutes = dct.get("timeoutInMinutes", "NA")
            queuedTimeoutInMinutes = dct.get("queuedTimeoutInMinutes", "NA")
            resourceAccessRole = dct.get("resourceAccessRole", "NA")
            source_type = dct.get("source", {}).get("type", "NA")
            source_location = dct.get("source", {}).get("location", "NA")
            source_buildspec = dct.get("source", {}).get("buildspec", "NA")
            environment_type = dct.get("environment", {}).get("type", "NA")
            environment_image = dct.get("environment", {}).get("image", "NA")
            environment_computeType = dct.get("environment", {}).get("computeType", "NA")
            environment_privilegedMode = dct.get("environment", {}).get("privilegedMode", "NA")
            env_vars = dct.get("environment", {}).get("environmentVariables", [])
            detail_items.extend([
                from_detail("description", description, url=url),
                from_detail("serviceRole", serviceRole, url=ars.aws_console.iam.get_role(serviceRole)),
                from_detail("concurrentBuildLimit", concurrentBuildLimit, url=url),
                from_detail("timeoutInMinutes", timeoutInMinutes, url=url),
                from_detail("queuedTimeoutInMinutes", queuedTimeoutInMinutes, url=url),
                from_detail("resourceAccessRole", resourceAccessRole, url=ars.aws_console.iam.get_role(serviceRole)),
                from_detail("source_type", source_type, url=url),
                from_detail("source_location", source_location, url=url),
                from_detail("source_buildspec", source_buildspec, value_text=self.one_line(source_buildspec), url=url),
                from_detail("environment_type", environment_type, url=url),
                from_detail("environment_image", environment_image, url=url),
                from_detail("environment_computeType", environment_computeType, url=url),
                from_detail("environment_privilegedMode", environment_privilegedMode, url=url),
            ])
            for d in env_vars:
                item = make_env_var_item(d["name"], d["value"], d["type"], url)
                detail_items.append(item)

            tags = rl.extract_tags(res)
            detail_items.extend(rl.DetailItem.from_tags(tags, url=url))

        return detail_items
    # fmt: on


class CodeBuildProjectSearcher(rl.BaseSearcher[CodeBuildProject]):
    pass


codebuild_project_searcher = CodeBuildProjectSearcher(
    # list resources
    service="codebuild",
    method="list_projects",
    is_paginator=True,
    default_boto_kwargs={
        "sortBy": "NAME",
        "sortOrder": "ASCENDING",
        "PaginationConfig": {
            "MaxItems": 5000,
        },
    },
    result_path=rl.ResultPath("projects"),
    # extract document
    doc_class=CodeBuildProject,
    # search
    resource_type=rl.SearcherEnum.codebuild_project.value,
    fields=CodeBuildProject.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.codebuild_project.value),
    more_cache_key=None,
)

codebuild_run_status_icon_mapper = {
    "SUCCEEDED": "🟢",
    "FAILED": "🔴",
    "FAULT": "🔴",
    "TIMED_OUT": "🔴",
    "IN_PROGRESS": "🟡",
    "STOPPED": "🔴",
}


@dataclasses.dataclass
class CodeBuildJobRun(rl.ResourceDocument):
    # fmt: off
    status: str = dataclasses.field(metadata={"field": sayt.NgramWordsField(name="status", minsize=2, maxsize=4, stored=True)})
    # fmt: on

    @property
    def fullname(self) -> str:
        return self.raw_data["id"]

    @property
    def short_id(self) -> int:
        return int(self.raw_data["id"].split(":", 1)[1])

    @property
    def start_at(self) -> datetime:
        return rl.get_datetime(self.raw_data, "startTime")

    @property
    def end_at(self) -> datetime:
        return rl.get_datetime(self.raw_data, "endTime")

    @classmethod
    def from_many_resources(
        cls,
        resources: rl.ResourceIterproxy,
        bsm,
        boto_kwargs,
    ):
        """
        We only pull the last 100 builds for each project.
        """
        ids = resources.all()[:100]
        if len(ids) == 0:
            return
        res = bsm.codebuild_client.batch_get_builds(ids=ids)
        for dct in res.get("builds", []):
            short_id = dct["id"].split(":", 1)[1]
            yield cls(
                raw_data=dct,
                id=short_id,
                name=short_id,
                status=dct["buildStatus"],
            )

    @property
    def title(self) -> str:
        return rl.format_key_value("build_id", self.id)

    @property
    def subtitle(self) -> str:
        utc_now = rl.get_utc_now()
        if self.raw_data.get("endTime"):
            duration = int((self.end_at - self.start_at).total_seconds())
            completed = int((utc_now - self.end_at).total_seconds())
            completed = rl.to_human_readable_elapsed(completed)
        else:
            duration = int((utc_now - self.start_at).total_seconds())
            completed = "Not yet"
        duration = rl.to_human_readable_elapsed(duration)
        status_icon = codebuild_run_status_icon_mapper[self.status]
        return "{}, {}, {}, {}".format(
            f"{status_icon} {self.status}",
            rl.format_key_value("duration", duration),
            rl.format_key_value("completed", completed),
            self.short_subtitle,
        )

    @property
    def autocomplete(self) -> str:
        project_name = self.fullname.split(":", 1)[0]
        return f"{project_name}@{self.id}"

    @property
    def arn(self) -> str:
        return self.raw_data["arn"]

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.codebuild.get_build_run(run_id_or_arn=self.arn)

    # fmt: off
    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        from_detail = rl.DetailItem.from_detail
        url = self.get_console_url(console=ars.aws_console)
        detail_items = rl.DetailItem.get_initial_detail_items(doc=self, ars=ars)

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.codebuild_client.batch_get_builds(ids=[self.fullname])
            builds = res.get("builds", [])
            if len(builds) == 0:
                return [
                    rl.DetailItem.new(
                        title="🚨 build job run not found, maybe it's deleted?",
                        subtitle=f"{rl.ShortcutEnum.ENTER} to verify in AWS Console",
                        url=url,
                    )
                ]
            build = builds[0]
            run = CodeBuildJobRun(raw_data=build, id=self.id, name=self.name, status=build["buildStatus"])
            status_icon = codebuild_run_status_icon_mapper[run.status]
            detail_items.extend([
                from_detail("status", run.status, value_text=f"{status_icon} {run.status}", url=url),
                from_detail("start_at", run.start_at, url=url),
                from_detail("end_at", run.end_at, url=url),
            ])

        serviceRole = self.raw_data.get("serviceRole", "NA")
        source_buildspec = self.raw_data.get("source", {}).get("buildspec", "NA")
        initiator = self.raw_data.get("initiator", "NA")
        detail_items.extend([
            from_detail("serviceRole", serviceRole, url=ars.aws_console.iam.get_role(serviceRole)),
            from_detail("source_buildspec", source_buildspec, value_text=self.one_line(source_buildspec), url=url),
            from_detail("initiator", initiator, url=url),
        ])

        env_vars = self.raw_data.get("environment", {}).get("environmentVariables", [])
        for d in env_vars:
            item = make_env_var_item(d["name"], d["value"], d["type"], url)
            detail_items.append(item)

        return detail_items
    # fmt: on


class CodeBuildJobRunSearcher(rl.BaseSearcher):
    pass


codebuild_job_run_searcher = CodeBuildJobRunSearcher(
    # list resources
    service="codebuild",
    method="list_builds_for_project",
    is_paginator=True,
    default_boto_kwargs={
        "PaginationConfig": {
            "MaxItems": 100,
        },
    },
    result_path=rl.ResultPath("ids"),
    # extract document
    doc_class=CodeBuildJobRun,
    # search
    resource_type=rl.SearcherEnum.codebuild_job_run.value,
    fields=CodeBuildJobRun.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.codebuild_job_run.value),
    more_cache_key=None,
)
